import asyncio
import datetime
import unittest
from unittest import mock

import discord
from discord.ext import commands
import freezegun

from lgrez import config, bdd
from lgrez.blocs import tools
from test import mock_discord



class TestToolsObjectsUtilities(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.tools utility functions (section 1)."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    def test__find_by_mention_or_name(self):
        """Unit tests for tools._find_by_mention_or_name function."""
        # def _find_by_mention_or_name(collec, val, pattern=None,
        #                              must_be_found=False, raiser=None)
        _find_by_mention_or_name = tools._find_by_mention_or_name
        names = ["ab", "dzk", "0077"]
        ids = [125, 1213135, 1218819]
        collec = [mock.Mock() for i in range(3)]
        for mk, name, id in zip(collec, names, ids):
            mk.configure_mock(id=id, name=name)

        # val False
        ret = _find_by_mention_or_name(collec, "")
        self.assertIsNone(ret)
        # val & must_be_found True
        with self.assertRaises(ValueError):
            _find_by_mention_or_name(collec, "", must_be_found=True)

        # val by name found
        ret = _find_by_mention_or_name(collec, "dzk")
        self.assertEqual(ret, collec[1])
        # val by name not found
        ret = _find_by_mention_or_name(collec, "dzK")
        self.assertIsNone(ret)
        # val by name not found & must_be_found True
        with self.assertRaises(ValueError):
            _find_by_mention_or_name(collec, "dzK", must_be_found=True)

        # val by pattern found
        ret = _find_by_mention_or_name(collec, "<1218819>",
                                       pattern=r"<([0-9]{7})>")
        self.assertEqual(ret, collec[2])
        # val by pattern not found
        ret = _find_by_mention_or_name(collec, "<1218820>",
                                       pattern=r"<([0-9]{7})>")
        self.assertIsNone(ret)
        # val by pattern not found & must_be_found True
        with self.assertRaises(ValueError):
            _find_by_mention_or_name(collec, "<1218820>", must_be_found=True,
                                     pattern=r"<([0-9]{7})>")


    @mock.patch("lgrez.blocs.tools._find_by_mention_or_name")
    def test_channel(self, fbn_patch):
        """Unit tests for tools.channel function."""
        # def channel(nom, must_be_found=True)
        channel = tools.channel
        nom = mock.Mock()
        # no arg
        res = channel(nom)
        fbn_patch.assert_called_once_with(
            config.guild.channels, nom, pattern="<#([0-9]{18})>",
            must_be_found=True, raiser=mock.ANY)
        self.assertEqual(res, fbn_patch.return_value)
        fbn_patch.reset_mock()
        # MBF False
        res = channel(nom, must_be_found=False)
        fbn_patch.assert_called_once_with(
            config.guild.channels, nom, pattern="<#([0-9]{18})>",
            must_be_found=False, raiser=mock.ANY)
        self.assertEqual(res, fbn_patch.return_value)


    @mock.patch("lgrez.blocs.tools._find_by_mention_or_name")
    def test_role(self, fbn_patch):
        """Unit tests for tools.role function."""
        # def role(nom, must_be_found=True)
        role = tools.role
        nom = mock.Mock()
        # no arg
        res = role(nom)
        fbn_patch.assert_called_once_with(
            config.guild.roles, nom, pattern="<@&([0-9]{18})>",
            must_be_found=True, raiser=mock.ANY)
        self.assertEqual(res, fbn_patch.return_value)
        fbn_patch.reset_mock()
        # MBF False
        res = role(nom, must_be_found=False)
        fbn_patch.assert_called_once_with(
            config.guild.roles, nom, pattern="<@&([0-9]{18})>",
            must_be_found=False, raiser=mock.ANY)
        self.assertEqual(res, fbn_patch.return_value)


    @mock.patch("lgrez.blocs.tools._find_by_mention_or_name")
    def test_member(self, fbn_patch):
        """Unit tests for tools.member function."""
        # def member(nom, must_be_found=True)
        member = tools.member
        nom = mock.Mock()
        # no arg
        res = member(nom)
        fbn_patch.assert_called_once_with(
            config.guild.members, nom, pattern="<@!([0-9]{18})>",
            must_be_found=True, raiser=mock.ANY)
        self.assertEqual(res, fbn_patch.return_value)
        fbn_patch.reset_mock()
        # MBF False
        res = member(nom, must_be_found=False)
        fbn_patch.assert_called_once_with(
            config.guild.members, nom, pattern="<@!([0-9]{18})>",
            must_be_found=False, raiser=mock.ANY)
        self.assertEqual(res, fbn_patch.return_value)


    @mock.patch("lgrez.blocs.tools._find_by_mention_or_name")
    def test_emoji(self, fbn_patch):
        """Unit tests for tools.emoji function."""
        # def emoji(nom, must_be_found=True)
        emoji = tools.emoji
        nom = mock.Mock()
        # no arg
        res = emoji(nom)
        fbn_patch.assert_called_once_with(
            config.guild.emojis, nom, pattern="<:.*:([0-9]{18})>",
            must_be_found=True, raiser=mock.ANY)
        self.assertEqual(res, fbn_patch.return_value)
        fbn_patch.reset_mock()
        # MBF False
        res = emoji(nom, must_be_found=False)
        fbn_patch.assert_called_once_with(
            config.guild.emojis, nom, pattern="<:.*:([0-9]{18})>",
            must_be_found=False, raiser=mock.ANY)
        self.assertEqual(res, fbn_patch.return_value)


    def test_mention_MJ(self):
        """Unit tests for tools.mention_MJ function."""
        # def mention_MJ(arg)
        mention_MJ = tools.mention_MJ
        _rmjn = config.Role.mj.name
        config.Role.mj.name = "bzzt"
        # arg = webhook
        arg = mock.Mock(discord.User)
        res = mention_MJ(arg)
        self.assertEqual(res, config.Role.mj.mention)
        # arg = Member < MJ
        arg = mock.Mock(discord.Member)
        arg.top_role.__ge__=lambda s, o: False
        res = mention_MJ(arg)
        self.assertEqual(res, config.Role.mj.mention)
        # arg = Member >= MJ
        arg = mock.Mock(discord.Member)
        arg.top_role.__ge__=lambda s, o: True
        res = mention_MJ(arg)
        self.assertEqual(res, "@bzzt")
        # arg = Context < MJ
        arg = mock.Mock(commands.Context, author=mock.Mock(discord.Member))
        arg.author.top_role.__ge__=lambda s, o: False
        res = mention_MJ(arg)
        self.assertEqual(res, config.Role.mj.mention)
        # arg = Context >= MJ
        arg.author.top_role.__ge__=lambda s, o: True
        res = mention_MJ(arg)
        self.assertEqual(res, "@bzzt")
        # reset
        config.Role.mj.name = _rmjn



class TestToolsDecorators(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.tools decorators (section 3)."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    async def test_mjs_only(self):
        """Unit tests for tools.mjs_only check."""
        # mjs_only = commands.check_any(...)
        # ==> mjs_only.predicate = coroutine ctx -> bool
        mjs_only = tools.mjs_only
        # webhook
        ctx = mock_discord.get_ctx()
        ctx.message.webhook_id = 12
        self.assertTrue(await mjs_only.predicate(ctx))
        # non-MJ
        ctx = mock_discord.get_ctx()
        ctx.author.roles = [config.Role.everyone, config.Role.joueur_en_vie]
        with self.assertRaises(commands.CheckAnyFailure):
            await mjs_only.predicate(ctx)
        # Rédacteur
        ctx.author.roles = [config.Role.everyone, config.Role.joueur_en_vie,
                            config.Role.redacteur]
        with self.assertRaises(commands.CheckAnyFailure):
            await mjs_only.predicate(ctx)
        # MJ
        ctx.author.roles = [config.Role.everyone, config.Role.mj]
        self.assertTrue(await mjs_only.predicate(ctx))


    async def test_mjs_et_redacteurs(self):
        """Unit tests for tools.mjs_et_redacteurs check."""
        # mjs_et_redacteurs = commands.check_any(...)
        # ==> mjs_et_redacteurs.predicate = coroutine ctx -> bool
        mjs_et_redacteurs = tools.mjs_et_redacteurs
        # webhook
        ctx = mock_discord.get_ctx()
        ctx.message.webhook_id = 12
        self.assertTrue(await mjs_et_redacteurs.predicate(ctx))
        # non-MJ
        ctx = mock_discord.get_ctx()
        ctx.author.roles = [config.Role.everyone, config.Role.joueur_en_vie]
        with self.assertRaises(commands.CheckAnyFailure):
            await mjs_et_redacteurs.predicate(ctx)
        # Rédacteur
        ctx.author.roles = [config.Role.everyone, config.Role.joueur_en_vie,
                            config.Role.redacteur]
        self.assertTrue(await mjs_et_redacteurs.predicate(ctx))
        # MJ
        ctx.author.roles = [config.Role.everyone, config.Role.mj]
        self.assertTrue(await mjs_et_redacteurs.predicate(ctx))


    async def test_joueurs_only(self):
        """Unit tests for tools.joueurs_only check."""
        # joueurs_only = commands.check_any(...)
        # ==> joueurs_only.predicate = coroutine ctx -> bool
        joueurs_only = tools.joueurs_only
        # webhook
        ctx = mock_discord.get_ctx()
        ctx.message.webhook_id = 12
        with self.assertRaises(commands.MissingAnyRole):
            await joueurs_only.predicate(ctx)
        # non-MJ
        ctx = mock_discord.get_ctx()
        ctx.author.roles = [config.Role.everyone, config.Role.joueur_en_vie]
        self.assertTrue(await joueurs_only.predicate(ctx))
        # Joueur mort
        ctx.author.roles = [config.Role.everyone, config.Role.joueur_mort]
        self.assertTrue(await joueurs_only.predicate(ctx))
        # MJ
        ctx.author.roles = [config.Role.everyone, config.Role.mj]
        with self.assertRaises(commands.MissingAnyRole):
            await joueurs_only.predicate(ctx)


    async def test_vivants_only(self):
        """Unit tests for tools.vivants_only check."""
        # vivants_only = commands.check_any(...)
        # ==> vivants_only.predicate = coroutine ctx -> bool
        vivants_only = tools.vivants_only
        # webhook
        ctx = mock_discord.get_ctx()
        ctx.message.webhook_id = 12
        with self.assertRaises(commands.MissingRole):
            await vivants_only.predicate(ctx)
        # non-MJ
        ctx = mock_discord.get_ctx()
        ctx.author.roles = [config.Role.everyone, config.Role.joueur_en_vie]
        self.assertTrue(await vivants_only.predicate(ctx))
        # Joueur mort
        ctx.author.roles = [config.Role.everyone, config.Role.joueur_mort]
        with self.assertRaises(commands.MissingRole):
            await vivants_only.predicate(ctx)
        # Joueur mort + rédacteur
        ctx.author.roles = [config.Role.everyone, config.Role.joueur_mort,
                            config.Role.redacteur]
        with self.assertRaises(commands.MissingRole):
            await vivants_only.predicate(ctx)
        # MJ
        ctx.author.roles = [config.Role.everyone, config.Role.mj]
        with self.assertRaises(commands.MissingRole):
            await vivants_only.predicate(ctx)


    @mock.patch("lgrez.bdd.Joueur.from_member")
    async def test_private(self, fm_patch):
        """Unit tests for tools.private decorator."""
        # def private(callback)
        private = tools.private
        callback = mock.AsyncMock()
        _pcp = config.private_chan_prefix
        config.private_chan_prefix = "pr€f-!x-"
        cog = mock.Mock()
        args = [mock.Mock() for i in range(5)]
        kwargs = {str(i): mock.Mock() for i in range(5)}

        # ok
        ctx = mock_discord.get_ctx()
        ctx.channel.name = "pr€f-!x-alloàtous"
        new = private(callback)
        await new(cog, ctx, *args, **kwargs)
        ctx.message.delete.assert_not_called()
        fm_patch.assert_not_called()
        ctx.assert_sent()
        callback.assert_called_once_with(cog, ctx, *args, **kwargs)
        callback.reset_mock()

        # not ok
        ctx = mock_discord.get_ctx()
        ctx.channel.name = "ono-alloàtous"
        ctx.message.content = "gzzzz"
        new = private(callback)
        await new(cog, ctx, *args, **kwargs)
        ctx.message.delete.assert_called_once()
        fm_patch.assert_called_once_with(ctx.author)
        ctx.assert_sent(["gzzzz", "interdite en dehors de ta conv privée"])
        self.assertEqual(ctx.channel, fm_patch.return_value.private_chan)
        callback.assert_called_once_with(cog, ctx, *args, **kwargs)

        # restore
        config.private_chan_prefix = _pcp



class TestToolsPlayerInteractors(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.tools player interactors (section 4)."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    async def test_wait_for_message(self):
        """Unit tests for tools.wait_for_message function."""
        # async def wait_for_message(check, trigger_on_commands=False)
        wait_for_message = tools.wait_for_message
        def primocheck(mess):
            return (mess.author != config.bot.user
                    and len(mess.content) > 4)

        # basic, pass
        config.bot.wait_for = mock.AsyncMock(return_value=mock.Mock(
            discord.Message, content="gloup"))
        mess = await wait_for_message(primocheck)
        config.bot.wait_for.assert_called_once_with("message", check=mock.ANY)
        check = config.bot.wait_for.call_args.kwargs["check"]
        self.assertTrue(check(mock.Mock(discord.Message, content="hmmooo")))
        self.assertFalse(check(mock.Mock(discord.Message, content="hmmm")))
        self.assertFalse(check(mock.Mock(discord.Message, content="hmmooo",
                                         author=config.bot.user)))
        self.assertTrue(check(mock.Mock(discord.Message, content="stop")))
        self.assertFalse(check(mock.Mock(discord.Message, content="!hmmo")))
        self.assertTrue(check(mock.Mock(discord.Message, content="!stop")))
        self.assertEqual(mess, config.bot.wait_for.return_value)

        # basic, stop, user
        config.bot.wait_for = mock.AsyncMock(return_value=mock.Mock(
            discord.Message, content="stop"))
        with self.assertRaises(tools.CommandExit) as cm:
            await wait_for_message(primocheck)
        config.bot.wait_for.assert_called_once_with("message", check=mock.ANY)
        self.assertIn("Arrêt demandé", cm.exception.args[0])

        # basic, stop, bot
        config.bot.wait_for = mock.AsyncMock(return_value=mock.Mock(
            discord.Message, content="stop", author=config.bot.user))
        with self.assertRaises(tools.CommandExit) as cm:
            await wait_for_message(primocheck)
        config.bot.wait_for.assert_called_once_with("message", check=mock.ANY)
        self.assertIn("Arrêt commande précédente", cm.exception.args[0])

        # toc True, pass
        config.bot.wait_for = mock.AsyncMock(return_value=mock.Mock(
            discord.Message, content="gloup"))
        mess = await wait_for_message(primocheck, trigger_on_commands=True)
        config.bot.wait_for.assert_called_once_with("message", check=mock.ANY)
        check = config.bot.wait_for.call_args.kwargs["check"]
        self.assertTrue(check(mock.Mock(discord.Message, content="hmmooo")))
        self.assertFalse(check(mock.Mock(discord.Message, content="hmmm")))
        self.assertFalse(check(mock.Mock(discord.Message, content="hmmooo",
                                         author=config.bot.user)))
        self.assertTrue(check(mock.Mock(discord.Message, content="stop")))
        self.assertTrue(check(mock.Mock(discord.Message, content="!hmmo")))
        self.assertTrue(check(mock.Mock(discord.Message, content="!stop")))
        self.assertEqual(mess, config.bot.wait_for.return_value)


    @mock.patch("lgrez.blocs.tools.wait_for_message")
    async def test_wait_for_message_here(self, wfm_patch):
        """Unit tests for tools.wait_for_message_here function."""
        # async def wait_for_message_here(ctx, trigger_on_commands=False)
        wait_for_message_here = tools.wait_for_message_here
        ctx = mock_discord.get_ctx()

        # basic
        wfm_patch.return_value = mock.Mock(discord.Message, content="gloup")
        mess = await wait_for_message_here(ctx)
        wfm_patch.assert_called_once_with(check=mock.ANY,
                                          trigger_on_commands=False)
        check = wfm_patch.call_args.kwargs["check"]
        self.assertFalse(check(mock.Mock(discord.Message, content="hmmooo")))
        self.assertTrue(check(mock.Mock(discord.Message, content="hmmooo",
                                        channel=ctx.channel)))
        self.assertFalse(check(mock.Mock(discord.Message, content="hmmooo",
                                         author=config.bot.user,
                                         channel=ctx.channel)))
        self.assertEqual(mess, wfm_patch.return_value)
        wfm_patch.reset_mock()

        # toc True
        wfm_patch.return_value = mock.Mock(discord.Message, content="gloup")
        mess = await wait_for_message_here(ctx, trigger_on_commands=True)
        wfm_patch.assert_called_once_with(check=mock.ANY,
                                          trigger_on_commands=True)
        check = wfm_patch.call_args.kwargs["check"]
        self.assertFalse(check(mock.Mock(discord.Message, content="hmmooo")))
        self.assertTrue(check(mock.Mock(discord.Message, content="hmmooo",
                                        channel=ctx.channel)))
        self.assertFalse(check(mock.Mock(discord.Message, content="hmmooo",
                                         author=config.bot.user,
                                         channel=ctx.channel)))
        self.assertEqual(mess, wfm_patch.return_value)


    @mock.patch("lgrez.blocs.tools.wait_for_message")
    async def test_boucle_message(self, wfm_patch):
        """Unit tests for tools.boucle_message function."""
        # async def boucle_message(chan, in_message, condition_sortie,
        #                          rep_message=None)
        boucle_message = tools.boucle_message

        # illegal
        with self.assertRaises(ValueError):
            await boucle_message(mock.Mock(), "", mock.Mock())

        # in_message only
        chan = mock_discord.chan(".")
        msgs = [mock.Mock(discord.Message, content="gloup1"),
                mock.Mock(discord.Message, content="gloup2"),
                mock.Mock(discord.Message, content="gloup3")]
        wfm_patch.side_effect = msgs
        condition_sortie = mock.Mock(side_effect=[False, False, True])
        mess = await boucle_message(chan, "in-m", condition_sortie)
        mock_discord.assert_sent(chan, "in-m", "in-m", "in-m")
        self.assertEqual(wfm_patch.call_count, 3)
        self.assertEqual(condition_sortie.call_count, 3)
        condition_sortie.assert_has_calls([mock.call(msg) for msg in msgs])
        check = wfm_patch.call_args.args[0]
        self.assertFalse(check(mock.Mock(discord.Message, content="hmmooo")))
        self.assertTrue(check(mock.Mock(discord.Message, content="hmmooo",
                                        channel=chan)))
        self.assertFalse(check(mock.Mock(discord.Message, content="hmmooo",
                                         author=config.bot.user,
                                         channel=chan)))
        self.assertEqual(mess, msgs[2])
        wfm_patch.reset_mock()

        # rep_message only
        chan = mock_discord.chan(".")
        msgs = [mock.Mock(discord.Message, content="gloup1"),
                mock.Mock(discord.Message, content="gloup2")]
        wfm_patch.side_effect = msgs
        condition_sortie = mock.Mock(side_effect=[False, True])
        mess = await boucle_message(chan, "", condition_sortie, "rep-m")
        mock_discord.assert_sent(chan, "rep-m")
        self.assertEqual(wfm_patch.call_count, 2)
        self.assertEqual(condition_sortie.call_count, 2)
        condition_sortie.assert_has_calls([mock.call(msg) for msg in msgs])
        self.assertEqual(mess, msgs[1])
        wfm_patch.reset_mock()

        # in_message and rep_message
        chan = mock_discord.chan(".")
        msgs = [mock.Mock(discord.Message, content="gloup1"),
                mock.Mock(discord.Message, content="gloup2"),
                mock.Mock(discord.Message, content="gloup3"),
                mock.Mock(discord.Message, content="gloup4")]
        wfm_patch.side_effect = msgs
        condition_sortie = mock.Mock(side_effect=[False, False, False, True])
        mess = await boucle_message(chan, "in-m", condition_sortie, "rep-m")
        mock_discord.assert_sent(chan, "in-m", "rep-m", "rep-m", "rep-m")
        self.assertEqual(wfm_patch.call_count, 4)
        self.assertEqual(condition_sortie.call_count, 4)
        condition_sortie.assert_has_calls([mock.call(msg) for msg in msgs])
        self.assertEqual(mess, msgs[3])
        wfm_patch.reset_mock()


    @mock.patch("lgrez.bdd.Joueur.from_member")
    @mock.patch("lgrez.bdd.Joueur.find_nearest")
    @mock.patch("lgrez.blocs.tools.member")
    async def test_boucle_query_joueur(self, mb_patch, find_patch, fm_patch):
        """Unit tests for tools.boucle_query_joueur function."""
        # async def boucle_query_joueur(ctx, cible=None, message=None,
        #                               sensi=0.5)
        boucle_query_joueur = tools.boucle_query_joueur
        ctx = mock_discord.get_ctx()

        # simple, first answer direct
        with mock_discord.interact(
            ("wait_for_message_here",
             mock_discord.message(ctx, "([)>>]}{<gl()zz()[}]{><)"))):
            jr = await boucle_query_joueur(ctx)
        mb_patch.assert_called_once_with("gl()zz", must_be_found=False)
        fm_patch.assert_called_once_with(mb_patch.return_value)
        self.assertEqual(jr, fm_patch.return_value)
        find_patch.assert_not_called()
        mb_patch.reset_mock()
        fm_patch.reset_mock()
        ctx.assert_sent()

        cible = mock.Mock()
        jr = await boucle_query_joueur(ctx, cible)
        mb_patch.assert_called_once_with(cible, must_be_found=False)
        fm_patch.assert_called_once_with(mb_patch.return_value)
        self.assertEqual(jr, fm_patch.return_value)
        find_patch.assert_not_called()
        mb_patch.reset_mock()
        fm_patch.reset_mock()
        ctx.assert_sent()

        # simple, first answer direct but not inscit -> fn exact
        fm_patch.side_effect = ValueError
        expjr = mock.Mock()
        find_patch.return_value = [(expjr, 1)]
        with mock_discord.interact(("wait_for_message_here",
                                    mock_discord.message(ctx, "<gloozz>"))):
            jr = await boucle_query_joueur(ctx)
        mb_patch.assert_called_once_with("gloozz", must_be_found=False)
        fm_patch.assert_called_once_with(mb_patch.return_value)
        find_patch.assert_called_once_with(
            "gloozz", col=bdd.Joueur.nom, sensi=0.5,
            solo_si_parfait=False, match_first_word=True)
        self.assertEqual(jr, expjr)
        find_patch.reset_mock()
        mb_patch.reset_mock()
        fm_patch.reset_mock()
        ctx.assert_sent()

        # cible given, first answer direct but not inscit -> fn exact (+seed)
        cible = mock.Mock()
        fm_patch.side_effect = ValueError
        expjr = mock.Mock()
        find_patch.return_value = [(expjr, 1)]
        jr = await boucle_query_joueur(ctx, cible, sensi=0.3)
        mb_patch.assert_called_once_with(cible, must_be_found=False)
        fm_patch.assert_called_once_with(mb_patch.return_value)
        find_patch.assert_called_once_with(
            cible, col=bdd.Joueur.nom, sensi=0.3,
            solo_si_parfait=False, match_first_word=True)
        self.assertEqual(jr, expjr)
        find_patch.reset_mock()
        mb_patch.reset_mock()
        fm_patch.reset_mock(side_effect=True)
        ctx.assert_sent()

        # simple, no member found -> fn approx, OK
        mb_patch.return_value = None
        expjr = mock.Mock(nom="boulip")
        find_patch.return_value = [(expjr, 0.8)]
        with mock_discord.interact(("wait_for_message_here",
                                    mock_discord.message(ctx, "<gloozz>")),
                ("yes_no", True)):
            jr = await boucle_query_joueur(ctx)
        mb_patch.assert_called_once_with("gloozz", must_be_found=False)
        find_patch.assert_called_once_with(
            "gloozz", col=bdd.Joueur.nom, sensi=0.5,
            solo_si_parfait=False, match_first_word=True)
        self.assertEqual(jr, expjr)
        fm_patch.assert_not_called()
        find_patch.reset_mock()
        mb_patch.reset_mock()
        ctx.assert_sent(["qu'une correspondance", "boulip"])
        ctx.reset_mock()

        # cible given, no member found -> fn approx, OK
        cible = mock.Mock()
        mb_patch.return_value = None
        expjr = mock.Mock(nom="boulip")
        find_patch.return_value = [(expjr, 0.8)]
        with mock_discord.interact(("yes_no", True)):
            jr = await boucle_query_joueur(ctx, cible)
        mb_patch.assert_called_once_with(cible, must_be_found=False)
        find_patch.assert_called_once_with(
            cible, col=bdd.Joueur.nom, sensi=0.5,
            solo_si_parfait=False, match_first_word=True)
        self.assertEqual(jr, expjr)
        fm_patch.assert_not_called()
        find_patch.reset_mock()
        mb_patch.reset_mock()
        ctx.assert_sent(["qu'une correspondance", "boulip"])
        ctx.reset_mock()


        # simple, no member found -> fn several, choice
        cible = mock.Mock()
        mb_patch.return_value = None
        expjrs = [mock.Mock(nom="boulip"), mock.Mock(nom="boul2"),
                  mock.Mock(nom="boo3"), mock.Mock(nom="bblbl4")]
        find_patch.return_value = list(zip(expjrs, [0.8, 0.8, 0.7, 0.5]))
        with mock_discord.interact(("wait_for_message_here",
                                    mock_discord.message(ctx, "<gloozz>")),
                ("choice", 3)) as int:
            jr = await boucle_query_joueur(ctx)
        mb_patch.assert_called_once_with("gloozz", must_be_found=False)
        find_patch.assert_called_once_with(
            "gloozz", col=bdd.Joueur.nom, sensi=0.5,
            solo_si_parfait=False, match_first_word=True)
        self.assertEqual(jr, expjrs[2])
        int.patchs["choice"].assert_called_with(config.bot, mock.ANY, 4)
        fm_patch.assert_not_called()
        find_patch.reset_mock()
        mb_patch.reset_mock()
        ctx.assert_sent(["joueurs les plus proches", f"1️⃣. boulip",
                         f"2️⃣. boul2", f"3️⃣. boo3", f"4️⃣. bblbl4",
                         "réagissant", "répondant"])
        ctx.reset_mock()

        # cible given, no member found -> fn several, choice (with > 10)
        cible = mock.Mock()
        mb_patch.return_value = None
        expjrs = [mock.Mock(nom=f"boul{i}") for i in range(15)]
        find_patch.return_value = list(zip(expjrs, [0.8]*15))
        with mock_discord.interact(("choice", 5)) as int:
            jr = await boucle_query_joueur(ctx, cible)
        mb_patch.assert_called_once_with(cible, must_be_found=False)
        find_patch.assert_called_once_with(
            cible, col=bdd.Joueur.nom, sensi=0.5,
            solo_si_parfait=False, match_first_word=True)
        self.assertEqual(jr, expjrs[4])
        int.patchs["choice"].assert_called_with(config.bot, mock.ANY, 10)
        fm_patch.assert_not_called()
        find_patch.reset_mock()
        mb_patch.reset_mock()
        ctx.assert_sent(["joueurs les plus proches", f"boul{0}",
                         f"🔟. boul9", "réagissant", "répondant"])
        ctx.reset_mock()


        # simple, no member found -> fn approx, DENY -> direct
        j2 = mock.Mock()
        mb_patch.side_effect = [None, j2]
        expjr = mock.Mock(nom="boulip")
        find_patch.return_value = [(expjr, 0.8)]
        with mock_discord.interact(("wait_for_message_here",
                                    mock_discord.message(ctx, "<gloozz>")),
             ("yes_no", False),
             ("wait_for_message_here", mock_discord.message(ctx, "<othy>"))):
            jr = await boucle_query_joueur(ctx)
        self.assertEqual(mb_patch.call_count, 2)
        mb_patch.assert_has_calls([
            mock.call("gloozz", must_be_found=False),
            mock.call("othy", must_be_found=False)])
        find_patch.assert_called_once_with(
            "gloozz", col=bdd.Joueur.nom, sensi=0.5,
            solo_si_parfait=False, match_first_word=True)
        fm_patch.assert_called_once_with(j2)
        self.assertEqual(jr, fm_patch.return_value)
        fm_patch.reset_mock()
        find_patch.reset_mock()
        mb_patch.reset_mock()
        ctx.assert_sent(["qu'une correspondance", "boulip"],
                         "d'accord, alors qui")
        ctx.reset_mock()

        # cible given, no member found -> fn approx, DENY -> direct
        cible = mock.Mock()
        j2 = mock.Mock()
        mb_patch.side_effect = [None, j2]
        expjr = mock.Mock(nom="boulip")
        find_patch.return_value = [(expjr, 0.8)]
        with mock_discord.interact(("yes_no", False),
             ("wait_for_message_here", mock_discord.message(ctx, "<othy>"))):
            jr = await boucle_query_joueur(ctx, cible)
        self.assertEqual(mb_patch.call_count, 2)
        mb_patch.assert_has_calls([
            mock.call(cible, must_be_found=False),
            mock.call("othy", must_be_found=False)])
        find_patch.assert_called_once_with(
            cible, col=bdd.Joueur.nom, sensi=0.5,
            solo_si_parfait=False, match_first_word=True)
        fm_patch.assert_called_once_with(j2)
        self.assertEqual(jr, fm_patch.return_value)
        fm_patch.reset_mock()
        find_patch.reset_mock()
        mb_patch.reset_mock()
        ctx.assert_sent(["qu'une correspondance", "boulip"],
                         "d'accord, alors qui")
        ctx.reset_mock()


        # simple, no member found -> fn none, redo -> direct
        j2 = mock.Mock()
        mb_patch.side_effect = [None, j2]
        find_patch.return_value = []
        with mock_discord.interact(("wait_for_message_here",
                                    mock_discord.message(ctx, "<gloozz>")),
             ("wait_for_message_here", mock_discord.message(ctx, "<othy>"))):
            jr = await boucle_query_joueur(ctx)
        self.assertEqual(mb_patch.call_count, 2)
        mb_patch.assert_has_calls([
            mock.call("gloozz", must_be_found=False),
            mock.call("othy", must_be_found=False)])
        find_patch.assert_called_once_with(
            "gloozz", col=bdd.Joueur.nom, sensi=0.5,
            solo_si_parfait=False, match_first_word=True)
        fm_patch.assert_called_once_with(j2)
        self.assertEqual(jr, fm_patch.return_value)
        fm_patch.reset_mock()
        find_patch.reset_mock()
        mb_patch.reset_mock()
        ctx.assert_sent(["Aucune entrée", "réessayer"])
        ctx.reset_mock()

        # cible given, no member found -> fn none -> direct
        cible = mock.Mock()
        j2 = mock.Mock()
        mb_patch.side_effect = [None, j2]
        find_patch.return_value = []
        with mock_discord.interact(
             ("wait_for_message_here", mock_discord.message(ctx, "<othy>"))):
            jr = await boucle_query_joueur(ctx, cible)
        self.assertEqual(mb_patch.call_count, 2)
        mb_patch.assert_has_calls([
            mock.call(cible, must_be_found=False),
            mock.call("othy", must_be_found=False)])
        find_patch.assert_called_once_with(
            cible, col=bdd.Joueur.nom, sensi=0.5,
            solo_si_parfait=False, match_first_word=True)
        fm_patch.assert_called_once_with(j2)
        self.assertEqual(jr, fm_patch.return_value)
        fm_patch.reset_mock()
        find_patch.reset_mock()
        mb_patch.reset_mock()
        ctx.assert_sent(["Aucune entrée", "réessayer"])
        ctx.reset_mock()

        # cible given, no member found -> fn none -> re none -> re -> -> direct
        cible = mock.Mock()
        j2 = mock.Mock()
        mb_patch.side_effect = [None, None, None, j2]
        find_patch.return_value = []
        with mock_discord.interact(
             ("wait_for_message_here", mock_discord.message(ctx, "<othy>")),
             ("wait_for_message_here", mock_discord.message(ctx, "(oula)")),
             ("wait_for_message_here", mock_discord.message(ctx, "{méssi}"))):
            jr = await boucle_query_joueur(ctx, cible)
        self.assertEqual(mb_patch.call_count, 4)
        mb_patch.assert_has_calls([
            mock.call(cible, must_be_found=False),
            mock.call("othy", must_be_found=False),
            mock.call("oula", must_be_found=False),
            mock.call("méssi", must_be_found=False)])
        self.assertEqual(find_patch.call_count, 3)
        find_patch.assert_has_calls([
            mock.call(cible, col=bdd.Joueur.nom, sensi=0.5,
                      solo_si_parfait=False, match_first_word=True),
            mock.call("othy", col=bdd.Joueur.nom, sensi=0.5,
                      solo_si_parfait=False, match_first_word=True),
            mock.call("oula", col=bdd.Joueur.nom, sensi=0.5,
                      solo_si_parfait=False, match_first_word=True)])
        fm_patch.assert_called_once_with(j2)
        self.assertEqual(jr, fm_patch.return_value)
        fm_patch.reset_mock()
        find_patch.reset_mock()
        mb_patch.reset_mock()
        ctx.assert_sent(["Aucune entrée", "réessayer"],
                        ["Aucune entrée", "réessayer"],
                        ["Aucune entrée", "réessayer"])
        ctx.reset_mock()

        # > 5 none
        cible = mock.Mock()
        mb_patch.side_effect = [None, None, None, None, None, None]
        find_patch.return_value = []
        with self.assertRaises(RuntimeError):
            with mock_discord.interact(
                 ("wait_for_message_here", mock_discord.message(ctx, "<ot>")),
                 ("wait_for_message_here", mock_discord.message(ctx, "(ou)")),
                 ("wait_for_message_here", mock_discord.message(ctx, "{mé}")),
                 ("wait_for_message_here", mock_discord.message(ctx, "[al]")),
                 ("wait_for_message_here", mock_discord.message(ctx, "po"))):
                jr = await boucle_query_joueur(ctx, cible)
        self.assertEqual(mb_patch.call_count, 5)
        self.assertEqual(find_patch.call_count, 5)
        fm_patch.assert_not_called()
        find_patch.reset_mock()
        mb_patch.reset_mock()
        ctx.assert_sent(["Aucune entrée", "réessayer"],
                        ["Aucune entrée", "réessayer"],
                        ["Aucune entrée", "réessayer"],
                        ["Aucune entrée", "réessayer"],
                        ["Aucune entrée", "réessayer"],
                        ["Et puis non"])
        ctx.reset_mock()


    @mock.patch("lgrez.blocs.tools.wait_for_message", new_callable=mock.Mock)
    @mock.patch("asyncio.create_task")
    @mock.patch("asyncio.wait")
    async def test_wait_for_react_clic(self, aw_patch, act_patch, wfm_patch):
        """Unit tests for tools.wait_for_react_clic function."""
        # async def wait_for_react_clic(message, emojis={}, *,
        #                               process_text=False, text_filter=None,
        #                               post_converter=None,
        #                               trigger_all_reacts=False,
        #                               trigger_on_commands=False)
        wait_for_react_clic = tools.wait_for_react_clic
        ctx = mock_discord.get_ctx()

        # simple, clic on react
        message = mock_discord.message(ctx)
        tasks = [mock.Mock(), mock.Mock()]      # react, message
        tasks[0].result.return_value.emoji = config.Emoji.lune
        act_patch.side_effect = tasks
        aw_patch.return_value = {tasks[0]}, {tasks[1]}

        ret = await wait_for_react_clic(message, {config.Emoji.lune: 1337,
                                                  "❤": 1565})
        config.bot.wait_for.assert_called_once_with('raw_reaction_add',
                                                    check=mock.ANY)
        wfm_patch.assert_called_once_with(trigger_on_commands=False,
                                          check=mock.ANY)
        react_check = config.bot.wait_for.call_args.kwargs["check"]
        message_check = wfm_patch.call_args.kwargs["check"]
        self.assertEqual(act_patch.call_count, 2)
        act_patch.assert_has_calls([mock.call(config.bot.wait_for.return_value),
                                    mock.call(wfm_patch.return_value)])
        aw_patch.assert_called_once_with(tasks,
                                         return_when=asyncio.FIRST_COMPLETED)
        self.assertEqual(message.remove_reaction.call_count, 2)
        message.remove_reaction.assert_has_calls([
            mock.call(config.Emoji.lune, config.bot.user),
            mock.call("❤", config.bot.user)])
        self.assertEqual(ret, 1337)
        aw_patch.reset_mock()
        act_patch.reset_mock()
        wfm_patch.reset_mock()

        # interlude: check checks
        message.id = 15
        config.bot.user.id = 33
        coeur = mock.Mock(discord.PartialEmoji)
        coeur.configure_mock(name="❤")
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=31,
                                               emoji=config.Emoji.maire)))
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=33,
                                               emoji=config.Emoji.maire)))
        self.assertFalse(react_check(mock.Mock(message_id=15, user_id=31,
                                               emoji=config.Emoji.maire)))
        self.assertFalse(react_check(mock.Mock(message_id=15, user_id=33,
                                               emoji=config.Emoji.maire)))
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=31,
                                               emoji=coeur)))
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=33,
                                               emoji=coeur)))
        self.assertTrue(react_check(mock.Mock(message_id=15, user_id=31,
                                               emoji=coeur)))
        self.assertFalse(react_check(mock.Mock(message_id=15, user_id=33,
                                               emoji=coeur)))
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=31,
                                               emoji=config.Emoji.lune)))
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=33,
                                               emoji=config.Emoji.lune)))
        self.assertTrue(react_check(mock.Mock(message_id=15, user_id=31,
                                               emoji=config.Emoji.lune)))
        self.assertFalse(react_check(mock.Mock(message_id=15, user_id=33,
                                               emoji=config.Emoji.lune)))
        chan = message.channel
        auth = config.bot.user
        self.assertFalse(message_check(mock.Mock(channel=chan, author="meh")))
        # messages not detected
        config.bot.wait_for.reset_mock()

        # message response
        message = mock_discord.message(ctx)
        tasks = [mock.Mock(), mock.Mock()]      # react, message
        tasks[1].result.return_value = mock.Mock(discord.Message,
                                                 content="alloz")
        act_patch.side_effect = tasks
        aw_patch.return_value = {tasks[1]}, {tasks[0]}

        ret = await wait_for_react_clic(message, {config.Emoji.lune: 1337},
                                        process_text=True)
        config.bot.wait_for.assert_called_once_with('raw_reaction_add',
                                                    check=mock.ANY)
        wfm_patch.assert_called_once_with(trigger_on_commands=False,
                                          check=mock.ANY)
        message_check = wfm_patch.call_args.kwargs["check"]
        self.assertEqual(act_patch.call_count, 2)
        act_patch.assert_has_calls(
            [mock.call(config.bot.wait_for.return_value),
             mock.call(wfm_patch.return_value)])
        aw_patch.assert_called_once_with(
            tasks, return_when=asyncio.FIRST_COMPLETED)
        mess = tasks[1].result.return_value
        message.clear_reactions.assert_called_once()
        self.assertEqual(ret, "alloz")
        aw_patch.reset_mock()
        act_patch.reset_mock()
        wfm_patch.reset_mock()
        config.bot.wait_for.reset_mock()

        # interlude: check check
        chan = message.channel
        auth = config.bot.user
        self.assertFalse(message_check(mock.Mock(channel="oh", author="meh")))
        self.assertFalse(message_check(mock.Mock(channel="oh", author=auth)))
        self.assertTrue(message_check(mock.Mock(channel=chan, author="meh")))
        self.assertFalse(message_check(mock.Mock(channel=chan, author=auth)))

        # simple, clic on react unicode
        message = mock_discord.message(ctx)
        tasks = [mock.Mock(), mock.Mock()]      # react, message
        tasks[0].result.return_value.emoji = "❤"
        act_patch.side_effect = tasks
        aw_patch.return_value = {tasks[0]}, {tasks[1]}

        ret = await wait_for_react_clic(message, {config.Emoji.lune: 1337,
                                                  "❤": 1565})
        config.bot.wait_for.assert_called_once_with('raw_reaction_add',
                                                    check=mock.ANY)
        wfm_patch.assert_called_once_with(trigger_on_commands=False,
                                          check=mock.ANY)
        react_check = config.bot.wait_for.call_args.kwargs["check"]
        message_check = wfm_patch.call_args.kwargs["check"]
        self.assertEqual(act_patch.call_count, 2)
        act_patch.assert_has_calls([mock.call(config.bot.wait_for.return_value),
                                    mock.call(wfm_patch.return_value)])
        aw_patch.assert_called_once_with(tasks,
                                         return_when=asyncio.FIRST_COMPLETED)
        self.assertEqual(message.remove_reaction.call_count, 2)
        message.remove_reaction.assert_has_calls([
            mock.call(config.Emoji.lune, config.bot.user),
            mock.call("❤", config.bot.user)])
        self.assertEqual(ret, 1565)
        aw_patch.reset_mock()
        act_patch.reset_mock()
        wfm_patch.reset_mock()
        config.bot.wait_for.reset_mock()

        # emoji list, clic on custom
        message = mock_discord.message(ctx)
        tasks = [mock.Mock(), mock.Mock()]      # react, message
        tasks[0].result.return_value.emoji = config.Emoji.lune
        act_patch.side_effect = tasks
        aw_patch.return_value = {tasks[0]}, {tasks[1]}

        ret = await wait_for_react_clic(message, [config.Emoji.lune, "❤"])
        config.bot.wait_for.assert_called_once_with('raw_reaction_add',
                                                    check=mock.ANY)
        wfm_patch.assert_called_once_with(trigger_on_commands=False,
                                          check=mock.ANY)
        react_check = config.bot.wait_for.call_args.kwargs["check"]
        message_check = wfm_patch.call_args.kwargs["check"]
        self.assertEqual(act_patch.call_count, 2)
        act_patch.assert_has_calls([mock.call(config.bot.wait_for.return_value),
                                    mock.call(wfm_patch.return_value)])
        aw_patch.assert_called_once_with(tasks,
                                         return_when=asyncio.FIRST_COMPLETED)
        self.assertEqual(message.remove_reaction.call_count, 2)
        message.remove_reaction.assert_has_calls([
            mock.call(config.Emoji.lune, config.bot.user),
            mock.call("❤", config.bot.user)])
        self.assertEqual(ret, config.Emoji.lune)
        aw_patch.reset_mock()
        act_patch.reset_mock()
        wfm_patch.reset_mock()
        config.bot.wait_for.reset_mock()

        # emoji list, clic on custom
        message = mock_discord.message(ctx)
        tasks = [mock.Mock(), mock.Mock()]      # react, message
        tasks[0].result.return_value.emoji = "❤"
        act_patch.side_effect = tasks
        aw_patch.return_value = {tasks[0]}, {tasks[1]}

        ret = await wait_for_react_clic(message, [config.Emoji.lune, "❤"])
        config.bot.wait_for.assert_called_once_with('raw_reaction_add',
                                                    check=mock.ANY)
        wfm_patch.assert_called_once_with(trigger_on_commands=False,
                                          check=mock.ANY)
        react_check = config.bot.wait_for.call_args.kwargs["check"]
        message_check = wfm_patch.call_args.kwargs["check"]
        self.assertEqual(act_patch.call_count, 2)
        act_patch.assert_has_calls([mock.call(config.bot.wait_for.return_value),
                                    mock.call(wfm_patch.return_value)])
        aw_patch.assert_called_once_with(tasks,
                                         return_when=asyncio.FIRST_COMPLETED)
        self.assertEqual(message.remove_reaction.call_count, 2)
        message.remove_reaction.assert_has_calls([
            mock.call(config.Emoji.lune, config.bot.user),
            mock.call("❤", config.bot.user)])
        self.assertEqual(ret, "❤")
        aw_patch.reset_mock()
        act_patch.reset_mock()
        wfm_patch.reset_mock()
        config.bot.wait_for.reset_mock()

        # trigger_all_reacts, clic on custom
        message = mock_discord.message(ctx)
        tasks = [mock.Mock(), mock.Mock()]      # react, message
        tasks[0].result.return_value.emoji = config.Emoji.maire
        act_patch.side_effect = tasks
        aw_patch.return_value = {tasks[0]}, {tasks[1]}

        ret = await wait_for_react_clic(message, [config.Emoji.lune, "❤"],
                                        trigger_all_reacts=True)
        config.bot.wait_for.assert_called_once_with('raw_reaction_add',
                                                    check=mock.ANY)
        wfm_patch.assert_called_once_with(trigger_on_commands=False,
                                          check=mock.ANY)
        react_check = config.bot.wait_for.call_args.kwargs["check"]
        self.assertEqual(act_patch.call_count, 2)
        act_patch.assert_has_calls([mock.call(config.bot.wait_for.return_value),
                                    mock.call(wfm_patch.return_value)])
        aw_patch.assert_called_once_with(tasks,
                                         return_when=asyncio.FIRST_COMPLETED)
        self.assertEqual(message.remove_reaction.call_count, 2)
        message.remove_reaction.assert_has_calls([
            mock.call(config.Emoji.lune, config.bot.user),
            mock.call("❤", config.bot.user)])
        self.assertEqual(ret, config.Emoji.maire)
        aw_patch.reset_mock()
        act_patch.reset_mock()
        wfm_patch.reset_mock()
        config.bot.wait_for.reset_mock()

        # interlude: check checks
        message.id = 15
        config.bot.user.id = 33
        coeur = mock.Mock(discord.PartialEmoji)
        coeur.configure_mock(name="❤")
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=31,
                                               emoji=config.Emoji.maire)))
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=33,
                                               emoji=config.Emoji.maire)))
        self.assertTrue(react_check(mock.Mock(message_id=15, user_id=31,
                                               emoji=config.Emoji.maire)))
        self.assertFalse(react_check(mock.Mock(message_id=15, user_id=33,
                                               emoji=config.Emoji.maire)))
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=31,
                                               emoji=coeur)))
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=33,
                                               emoji=coeur)))
        self.assertTrue(react_check(mock.Mock(message_id=15, user_id=31,
                                               emoji=coeur)))
        self.assertFalse(react_check(mock.Mock(message_id=15, user_id=33,
                                               emoji=coeur)))
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=31,
                                               emoji=config.Emoji.lune)))
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=33,
                                               emoji=config.Emoji.lune)))
        self.assertTrue(react_check(mock.Mock(message_id=15, user_id=31,
                                               emoji=config.Emoji.lune)))
        self.assertFalse(react_check(mock.Mock(message_id=15, user_id=33,
                                               emoji=config.Emoji.lune)))

        # message response - text_filter len >= 5 (and trigger_on_commands)
        message = mock_discord.message(ctx)
        tasks = [mock.Mock(), mock.Mock()]      # react, message
        tasks[1].result.return_value = mock.Mock(discord.Message,
                                                 content="alloz")
        act_patch.side_effect = tasks
        aw_patch.return_value = {tasks[1]}, {tasks[0]}

        ret = await wait_for_react_clic(
            message, {config.Emoji.lune: 1337}, process_text=True,
            text_filter=lambda s: len(s) >= 5, trigger_on_commands=True)
        config.bot.wait_for.assert_called_once_with('raw_reaction_add',
                                                    check=mock.ANY)
        wfm_patch.assert_called_once_with(trigger_on_commands=True,
                                          check=mock.ANY)
        message_check = wfm_patch.call_args.kwargs["check"]
        self.assertEqual(act_patch.call_count, 2)
        act_patch.assert_has_calls(
            [mock.call(config.bot.wait_for.return_value),
             mock.call(wfm_patch.return_value)])
        aw_patch.assert_called_once_with(
            tasks, return_when=asyncio.FIRST_COMPLETED)
        mess = tasks[1].result.return_value
        message.clear_reactions.assert_called_once()
        self.assertEqual(ret, "alloz")
        aw_patch.reset_mock()
        act_patch.reset_mock()
        wfm_patch.reset_mock()
        config.bot.wait_for.reset_mock()
        # check check
        chan = message.channel
        auth = config.bot.user
        self.assertFalse(message_check(mock.Mock(channel="oh", author="meh",
                                                 content="ala")))
        self.assertFalse(message_check(mock.Mock(channel="oh", author=auth,
                                                 content="ala")))
        self.assertFalse(message_check(mock.Mock(channel=chan, author="meh",
                                                 content="ala")))
        self.assertFalse(message_check(mock.Mock(channel=chan, author=auth,
                                                 content="ala")))
        self.assertFalse(message_check(mock.Mock(channel="oh", author="meh",
                                                 content="alala")))
        self.assertFalse(message_check(mock.Mock(channel="oh", author=auth,
                                                 content="alala")))
        self.assertTrue(message_check(mock.Mock(channel=chan, author="meh",
                                                 content="alala")))
        self.assertFalse(message_check(mock.Mock(channel=chan, author=auth,
                                                 content="alala")))

        # message response - post_converter
        message = mock_discord.message(ctx)
        tasks = [mock.Mock(), mock.Mock()]      # react, message
        tasks[1].result.return_value = mock.Mock(discord.Message,
                                                 content="alloz")
        act_patch.side_effect = tasks
        aw_patch.return_value = {tasks[1]}, {tasks[0]}
        ret = await wait_for_react_clic(
            message, {config.Emoji.lune: 1337}, process_text=True,
            post_converter=lambda s: hash(s) + 19.7)
        config.bot.wait_for.assert_called_once_with('raw_reaction_add',
                                                    check=mock.ANY)
        wfm_patch.assert_called_once_with(trigger_on_commands=False,
                                          check=mock.ANY)
        message_check = wfm_patch.call_args.kwargs["check"]
        self.assertEqual(act_patch.call_count, 2)
        act_patch.assert_has_calls(
            [mock.call(config.bot.wait_for.return_value),
             mock.call(wfm_patch.return_value)])
        aw_patch.assert_called_once_with(
            tasks, return_when=asyncio.FIRST_COMPLETED)
        mess = tasks[1].result.return_value
        message.clear_reactions.assert_called_once()
        self.assertEqual(ret, hash("alloz") + 19.7)
        aw_patch.reset_mock()
        act_patch.reset_mock()
        wfm_patch.reset_mock()
        config.bot.wait_for.reset_mock()

        # exception raised
        message = mock_discord.message(ctx)
        config.bot.wait_for.side_effect = RuntimeError
        with self.assertRaises(RuntimeError):
            ret = await wait_for_react_clic(message, {config.Emoji.lune: 1337})
        message.clear_reactions.assert_called_once()
        aw_patch.assert_not_called()
        act_patch.assert_not_called()
        wfm_patch.assert_not_called()
        config.bot.wait_for.assert_called_once()
        config.bot.wait_for.reset_mock()


    @mock.patch("lgrez.blocs.tools.wait_for_react_clic")
    async def test_yes_no(self, wfrc_patch):
        """Unit tests for tools.yes_no function."""
        # async def yes_no(message)
        yes_no = tools.yes_no

        # single case
        message = mock.Mock()
        ret = await yes_no(message)
        wfrc_patch.assert_called_once_with(
            message, emojis={"✅": True, "❎": False}, process_text=True,
            text_filter=mock.ANY, post_converter=mock.ANY
        )
        text_filter = wfrc_patch.call_args.kwargs["text_filter"]
        self.assertTrue(text_filter("oui"))
        self.assertTrue(text_filter("OUI"))
        self.assertTrue(text_filter("OuI"))
        self.assertTrue(text_filter("yes"))
        self.assertTrue(text_filter("YeS"))
        self.assertTrue(text_filter("1"))
        self.assertTrue(text_filter("non"))
        self.assertTrue(text_filter("NON"))
        self.assertTrue(text_filter("NoN"))
        self.assertTrue(text_filter("no"))
        self.assertTrue(text_filter("NO"))
        self.assertTrue(text_filter("0"))
        self.assertFalse(text_filter("."))
        self.assertFalse(text_filter("ye"))
        self.assertFalse(text_filter("ou"))
        self.assertFalse(text_filter("bloup"))
        self.assertFalse(text_filter("2"))

        post_converter = wfrc_patch.call_args.kwargs["post_converter"]
        self.assertTrue(post_converter("oui"))
        self.assertTrue(post_converter("OUI"))
        self.assertTrue(post_converter("OuI"))
        self.assertTrue(post_converter("yes"))
        self.assertTrue(post_converter("YeS"))
        self.assertTrue(post_converter("1"))
        self.assertFalse(post_converter("non"))
        self.assertFalse(post_converter("NON"))
        self.assertFalse(post_converter("NoN"))
        self.assertFalse(post_converter("no"))
        self.assertFalse(post_converter("NO"))
        self.assertFalse(post_converter("0"))


    @mock.patch("lgrez.blocs.tools.wait_for_react_clic")
    async def test_choice(self, wfrc_patch):
        """Unit tests for tools.choice function."""
        # async def choice(message, N, start=1, *, additionnal={})
        choice = tools.choice

        # simple
        message = mock.Mock()
        ret = await choice(message, 3)
        wfrc_patch.assert_called_once_with(
            message, emojis={"1️⃣": 1, "2️⃣": 2, "3️⃣": 3}, process_text=True,
            text_filter=mock.ANY, post_converter=int
        )
        text_filter = wfrc_patch.call_args.kwargs["text_filter"]
        self.assertFalse(text_filter("0"))
        self.assertTrue(text_filter("1"))
        self.assertTrue(text_filter("2"))
        self.assertTrue(text_filter("3"))
        self.assertFalse(text_filter("4"))
        self.assertFalse(text_filter("."))
        self.assertFalse(text_filter("yes"))
        self.assertFalse(text_filter("ou"))
        self.assertFalse(text_filter("bloup"))
        wfrc_patch.reset_mock()

        # with start
        message = mock.Mock()
        ret = await choice(message, 7, start=4)
        wfrc_patch.assert_called_once_with(
            message, emojis={"4️⃣": 4, "5️⃣": 5, "6️⃣": 6, "7️⃣": 7},
            process_text=True, text_filter=mock.ANY, post_converter=int
        )
        text_filter = wfrc_patch.call_args.kwargs["text_filter"]
        self.assertFalse(text_filter("0"))
        self.assertFalse(text_filter("1"))
        self.assertFalse(text_filter("2"))
        self.assertFalse(text_filter("3"))
        self.assertTrue(text_filter("4"))
        self.assertTrue(text_filter("5"))
        self.assertTrue(text_filter("6"))
        self.assertTrue(text_filter("7"))
        self.assertFalse(text_filter("8"))
        self.assertFalse(text_filter("yes"))
        wfrc_patch.reset_mock()

        # with start and aditionnal
        message = mock.Mock()
        ret = await choice(message, 7, start=4,
                           additionnal={config.Emoji.lune: 1337, "❤": "zebi"})
        wfrc_patch.assert_called_once_with(
            message, emojis={"4️⃣": 4, "5️⃣": 5, "6️⃣": 6, "7️⃣": 7,
                             config.Emoji.lune: 1337, "❤": "zebi"},
            process_text=True, text_filter=mock.ANY, post_converter=int
        )
        text_filter = wfrc_patch.call_args.kwargs["text_filter"]
        self.assertFalse(text_filter("0"))
        self.assertFalse(text_filter("1"))
        self.assertFalse(text_filter("2"))
        self.assertFalse(text_filter("3"))
        self.assertTrue(text_filter("4"))
        self.assertTrue(text_filter("5"))
        self.assertTrue(text_filter("6"))
        self.assertTrue(text_filter("7"))
        self.assertFalse(text_filter("8"))
        self.assertFalse(text_filter("yes"))


    @mock.patch("asyncio.sleep")
    async def test_sleep(self, sleep_patch):
        """Unit tests for tools.sleep function."""
        # async def sleep(chan, tps)
        sleep = tools.sleep
        chan = mock_discord.chan(".")
        await sleep(chan, 17.5)
        chan.typing.assert_called_once()
        sleep_patch.assert_called_once_with(17.5)



class TestToolsEmojisUtilities(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.tools emoji utility functions (section 5)."""

    def test_montre(self):
        """Unit tests for tools.montre function."""
        # def montre(heure=None)
        montre = tools.montre

        samples = {
            "kokf": ValueError,
            "ahb": ValueError,
            "7ha": ValueError,
            "7h": "\N{CLOCK FACE SEVEN OCLOCK}",
            "13h": "\N{CLOCK FACE ONE OCLOCK}",
            "0h": "\N{CLOCK FACE TWELVE OCLOCK}",
            "7h14": "\N{CLOCK FACE SEVEN OCLOCK}",
            "7h15": "\N{CLOCK FACE SEVEN OCLOCK}",
            "7h16": "\N{CLOCK FACE SEVEN-THIRTY}",
            "7h44": "\N{CLOCK FACE SEVEN-THIRTY}",
            "7h45": "\N{CLOCK FACE EIGHT OCLOCK}",
            "7h46": "\N{CLOCK FACE EIGHT OCLOCK}",
            "7h59": "\N{CLOCK FACE EIGHT OCLOCK}",
            "19h14": "\N{CLOCK FACE SEVEN OCLOCK}",
            "19h15": "\N{CLOCK FACE SEVEN OCLOCK}",
            "19h16": "\N{CLOCK FACE SEVEN-THIRTY}",
            "19h44": "\N{CLOCK FACE SEVEN-THIRTY}",
            "19h45": "\N{CLOCK FACE EIGHT OCLOCK}",
            "19h46": "\N{CLOCK FACE EIGHT OCLOCK}",
            "19h59": "\N{CLOCK FACE EIGHT OCLOCK}",
            "23h59": "\N{CLOCK FACE TWELVE OCLOCK}",
        }
        for sample, result in samples.items():
            if isinstance(result, type) and issubclass(result, Exception):
                with self.assertRaises(result):
                    montre(sample)
            else:
                self.assertEqual(montre(sample), result, msg=sample)
                with freezegun.freeze_time(sample):
                    self.assertEqual(montre(), result, msg=sample)


    def test_emoji_chiffre(self):
        """Unit tests for tools.emoji_chiffre function."""
        # def emoji_chiffre(chiffre, multi=False)
        emoji_chiffre = tools.emoji_chiffre

        samples = {
            -1: ValueError,
            11: ValueError,
            7.5: ValueError,
            0: "0️⃣",
            1: "1️⃣",
            2: "2️⃣",
            3: "3️⃣",
            4: "4️⃣",
            5: "5️⃣",
            6: "6️⃣",
            7: "7️⃣",
            8: "8️⃣",
            9: "9️⃣",
            10: "🔟",
        }
        samples_multionly = {
            72: "7️⃣2️⃣",
            75334: "7️⃣5️⃣3️⃣3️⃣4️⃣",
            100000: "1️⃣0️⃣0️⃣0️⃣0️⃣0️⃣",
        }
        for sample, result in samples.items():
            if isinstance(result, type) and issubclass(result, Exception):
                with self.assertRaises(result):
                    emoji_chiffre(sample)
            else:
                self.assertEqual(emoji_chiffre(sample), result, msg=sample)

        for sample, result in samples_multionly.items():
            with self.assertRaises(ValueError):
                emoji_chiffre(sample)
            self.assertEqual(emoji_chiffre(sample, multi=True), result,
                             msg=sample)


    def test_super_chiffre(self):
        """Unit tests for tools.super_chiffre function."""
        # def super_chiffre(chiffre, multi=False)
        super_chiffre = tools.super_chiffre

        samples = {
            -1: ValueError,
            11: ValueError,
            7.5: ValueError,
            0: "⁰",
            1: "¹",
            2: "²",
            3: "³",
            4: "⁴",
            5: "⁵",
            6: "⁶",
            7: "⁷",
            8: "⁸",
            9: "⁹",
        }
        samples_multionly = {
            10: "¹⁰",
            72: "⁷²",
            75334: "⁷⁵³³⁴",
            100000: "¹⁰⁰⁰⁰⁰",
        }
        for sample, result in samples.items():
            if isinstance(result, type) and issubclass(result, Exception):
                with self.assertRaises(result):
                    super_chiffre(sample)
            else:
                self.assertEqual(super_chiffre(sample), result, msg=sample)

        for sample, result in samples_multionly.items():
            with self.assertRaises(ValueError):
                super_chiffre(sample)
            self.assertEqual(super_chiffre(sample, multi=True), result,
                             msg=sample)


    def test_sub_chiffre(self):
        """Unit tests for tools.sub_chiffre function."""
        # def sub_chiffre(chiffre, multi=False)
        sub_chiffre = tools.sub_chiffre

        samples = {
            -1: ValueError,
            11: ValueError,
            7.5: ValueError,
            0: "₀",
            1: "₁",
            2: "₂",
            3: "₃",
            4: "₄",
            5: "₅",
            6: "₆",
            7: "₇",
            8: "₈",
            9: "₉",
        }
        samples_multionly = {
            10: "₁₀",
            72: "₇₂",
            75334: "₇₅₃₃₄",
            100000: "₁₀₀₀₀₀",
        }
        for sample, result in samples.items():
            if isinstance(result, type) and issubclass(result, Exception):
                with self.assertRaises(result):
                    sub_chiffre(sample)
            else:
                self.assertEqual(sub_chiffre(sample), result, msg=sample)

        for sample, result in samples_multionly.items():
            with self.assertRaises(ValueError):
                sub_chiffre(sample)
            self.assertEqual(sub_chiffre(sample, multi=True), result,
                             msg=sample)



class TestToolsTimeUtilities(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.tools time utility functions (section 6)."""

    def test_heure_to_time(self):
        """Unit tests for tools.heure_to_time function."""
        # def heure_to_time(heure=None)
        heure_to_time = tools.heure_to_time

        samples = {
            "kokf": ValueError,
            "ahb": ValueError,
            "7ha": ValueError,
            "h2": ValueError,
            "7h": datetime.time(7, 0),
            "13h": datetime.time(13, 0),
            "0h": datetime.time(0, 0),
            "0h0": datetime.time(0, 0),
            "0h00": datetime.time(0, 0),
            "00h00": datetime.time(0, 0),
            "7h14": datetime.time(7, 14),
            "07h46": datetime.time(7, 46),
            "7h59": datetime.time(7, 59),
            "19h04": datetime.time(19, 4),
            "19h5": datetime.time(19, 5),
            "19h59": datetime.time(19, 59),
            "23h59": datetime.time(23, 59),
            "7h60": ValueError,
            "24h12": ValueError,
            "-1h2": ValueError,
            "a:b": ValueError,
            "7:a": ValueError,
            ":2": ValueError,
            "7:": datetime.time(7, 0),
            "13:": datetime.time(13, 0),
            "0:": datetime.time(0, 0),
            "0:0": datetime.time(0, 0),
            "0:00": datetime.time(0, 0),
            "00:00": datetime.time(0, 0),
            "7:14": datetime.time(7, 14),
            "07:46": datetime.time(7, 46),
            "7:59": datetime.time(7, 59),
            "19:04": datetime.time(19, 4),
            "19:5": datetime.time(19, 5),
            "19:59": datetime.time(19, 59),
            "23:59": datetime.time(23, 59),
            "7:60": ValueError,
            "24:12": ValueError,
            "-1:2": ValueError,
        }
        for sample, result in samples.items():
            if isinstance(result, type) and issubclass(result, Exception):
                with self.assertRaises(result):
                    heure_to_time(sample)
            else:
                self.assertEqual(heure_to_time(sample), result, msg=sample)


    def test_time_to_heure(self):
        """Unit tests for tools.time_to_heure function."""
        # def time_to_heure(tps, sep="h", force_minutes=False)
        time_to_heure = tools.time_to_heure

        samples = {
            datetime.time(7, 0): ("7h", "7h00", "7<S€>", "7<S€>00"),
            datetime.time(13, 0): ("13h", "13h00", "13<S€>", "13<S€>00"),
            datetime.time(0, 0): ("0h", "0h00", "0<S€>", "0<S€>00"),
            datetime.time(7, 14): ("7h14", "7h14", "7<S€>14", "7<S€>14"),
            datetime.time(7, 46): ("7h46", "7h46", "7<S€>46", "7<S€>46"),
            datetime.time(7, 59): ("7h59", "7h59", "7<S€>59", "7<S€>59"),
            datetime.time(19, 4): ("19h04", "19h04", "19<S€>04", "19<S€>04"),
            datetime.time(19, 59): ("19h59", "19h59", "19<S€>59", "19<S€>59"),
            datetime.time(23, 59): ("23h59", "23h59", "23<S€>59", "23<S€>59"),
        }

        for sp, expected in samples.items():
            results = [
                time_to_heure(sp), time_to_heure(sp, force_minutes=True),
                time_to_heure(sp, sep="<S€>"),
                time_to_heure(sp, sep="<S€>", force_minutes=True),
            ]
            self.assertEqual(list(expected), results, msg=sp)


    def test_debut_pause(self):
        """Unit tests for tools.debut_pause function."""
        # def debut_pause()
        debut_pause = tools.debut_pause
        # on se place la semaine du 1 mars 2021, parce que lundi = 1

        res1 = datetime.datetime(2021, 3, 5, 19, 0)     # vendredi 19h
        res2 = datetime.datetime(2021, 3, 12, 19, 0)    # semaine d'après

        samples1 = [
            datetime.datetime(2021, 3, 1, 0, 0),
            datetime.datetime(2021, 3, 1, 12, 15),
            datetime.datetime(2021, 3, 1, 23, 59),
            datetime.datetime(2021, 3, 2, 0, 0),
            datetime.datetime(2021, 3, 2, 12, 15),
            datetime.datetime(2021, 3, 2, 23, 59),
            datetime.datetime(2021, 3, 3, 0, 0),
            datetime.datetime(2021, 3, 3, 12, 15),
            datetime.datetime(2021, 3, 3, 23, 59),
            datetime.datetime(2021, 3, 4, 0, 0),
            datetime.datetime(2021, 3, 4, 12, 15),
            datetime.datetime(2021, 3, 4, 23, 59),
            datetime.datetime(2021, 3, 5, 0, 0),
            datetime.datetime(2021, 3, 5, 12, 15),
            datetime.datetime(2021, 3, 5, 18, 59),
            datetime.datetime(2021, 3, 5, 19, 0),
        ]
        samples2 = [
            datetime.datetime(2021, 3, 5, 19, 0, 1),
            datetime.datetime(2021, 3, 5, 19, 1),
            datetime.datetime(2021, 3, 5, 23, 59),
            datetime.datetime(2021, 3, 6, 0, 0),
            datetime.datetime(2021, 3, 6, 12, 15),
            datetime.datetime(2021, 3, 6, 23, 59),
            datetime.datetime(2021, 3, 7, 0, 0),
            datetime.datetime(2021, 3, 7, 12, 15),
            datetime.datetime(2021, 3, 7, 23, 59),
        ]

        for sample in samples1:
            with freezegun.freeze_time(sample):
                self.assertEqual(debut_pause(), res1, msg=sample)
        for sample in samples2:
            with freezegun.freeze_time(sample):
                self.assertEqual(debut_pause(), res2, msg=sample)


    def test_fin_pause(self):
        """Unit tests for tools.fin_pause function."""
        # def fin_pause()
        fin_pause = tools.fin_pause
        # on se place la semaine du 1 mars 2021, parce que lundi = 1

        res1 = datetime.datetime(2021, 3, 7, 19, 0)     # dimanche 19h
        res2 = datetime.datetime(2021, 3, 14, 19, 0)    # semaine d'après

        samples1 = [
            datetime.datetime(2021, 3, 1, 0, 0),
            datetime.datetime(2021, 3, 1, 12, 15),
            datetime.datetime(2021, 3, 1, 23, 59),
            datetime.datetime(2021, 3, 2, 0, 0),
            datetime.datetime(2021, 3, 2, 12, 15),
            datetime.datetime(2021, 3, 2, 23, 59),
            datetime.datetime(2021, 3, 3, 0, 0),
            datetime.datetime(2021, 3, 3, 12, 15),
            datetime.datetime(2021, 3, 3, 23, 59),
            datetime.datetime(2021, 3, 4, 0, 0),
            datetime.datetime(2021, 3, 4, 12, 15),
            datetime.datetime(2021, 3, 4, 23, 59),
            datetime.datetime(2021, 3, 5, 0, 0),
            datetime.datetime(2021, 3, 5, 12, 15),
            datetime.datetime(2021, 3, 5, 23, 59),
            datetime.datetime(2021, 3, 6, 0, 0),
            datetime.datetime(2021, 3, 6, 12, 15),
            datetime.datetime(2021, 3, 6, 23, 59),
            datetime.datetime(2021, 3, 7, 0, 0),
            datetime.datetime(2021, 3, 7, 12, 15),
            datetime.datetime(2021, 3, 7, 18, 59),
            datetime.datetime(2021, 3, 7, 19, 0),
        ]
        samples2 = [
            datetime.datetime(2021, 3, 7, 19, 0, 1),
            datetime.datetime(2021, 3, 7, 19, 1),
            datetime.datetime(2021, 3, 7, 23, 59),
        ]

        for sample in samples1:
            with freezegun.freeze_time(sample):
                self.assertEqual(fin_pause(), res1, msg=sample)
        for sample in samples2:
            with freezegun.freeze_time(sample):
                self.assertEqual(fin_pause(), res2, msg=sample)


    @mock.patch("lgrez.blocs.tools.debut_pause")
    @mock.patch("lgrez.blocs.tools.fin_pause")
    def test_en_pause(self, fin_patch, debut_patch):
        """Unit tests for tools.en_pause function."""
        # def en_pause()
        en_pause = tools.en_pause
        # oui
        fin_patch.return_value = 1
        debut_patch.return_value = 2
        self.assertTrue(en_pause())
        # non
        fin_patch.return_value = 3
        debut_patch.return_value = 2
        self.assertFalse(en_pause())
        # equal -> False
        fin_patch.return_value = 2
        debut_patch.return_value = 2
        self.assertFalse(en_pause())


    def test_next_occurence(self):
        """Unit tests for tools.next_occurence function."""
        # def next_occurence(tps)
        next_occurence = tools.next_occurence
        # on se place la semaine du 1 mars 2021, parce que lundi = 1

        samples = {}
        for i in range(1, 4):       # identique lundi -> mercredi
            samples.update({
            (datetime.datetime(2021, 3, i, 0, 0),   datetime.time(10, 0)): 0,
            (datetime.datetime(2021, 3, i, 0, 0),   datetime.time(14, 0)): 0,
            (datetime.datetime(2021, 3, i, 0, 0),   datetime.time(19, 0)): 0,
            (datetime.datetime(2021, 3, i, 0, 0),   datetime.time(22, 0)): 0,
            (datetime.datetime(2021, 3, i, 12, 15), datetime.time(10, 0)): 1,
            (datetime.datetime(2021, 3, i, 12, 15), datetime.time(14, 0)): 0,
            (datetime.datetime(2021, 3, i, 12, 15), datetime.time(19, 0)): 0,
            (datetime.datetime(2021, 3, i, 12, 15), datetime.time(22, 0)): 0,
            (datetime.datetime(2021, 3, i, 23, 59), datetime.time(10, 0)): 1,
            (datetime.datetime(2021, 3, i, 23, 59), datetime.time(14, 0)): 1,
            (datetime.datetime(2021, 3, i, 23, 59), datetime.time(19, 0)): 1,
            (datetime.datetime(2021, 3, i, 23, 59), datetime.time(22, 0)): 1,
            })
        samples.update({
            (datetime.datetime(2021, 3, 4, 0, 0),   datetime.time(10, 0)): 0,
            (datetime.datetime(2021, 3, 4, 0, 0),   datetime.time(14, 0)): 0,
            (datetime.datetime(2021, 3, 4, 0, 0),   datetime.time(19, 0)): 0,
            (datetime.datetime(2021, 3, 4, 0, 0),   datetime.time(22, 0)): 0,
            (datetime.datetime(2021, 3, 4, 12, 15), datetime.time(10, 0)): 1,
            (datetime.datetime(2021, 3, 4, 12, 15), datetime.time(14, 0)): 0,
            (datetime.datetime(2021, 3, 4, 12, 15), datetime.time(19, 0)): 0,
            (datetime.datetime(2021, 3, 4, 12, 15), datetime.time(22, 0)): 0,
            (datetime.datetime(2021, 3, 4, 23, 59), datetime.time(10, 0)): 1,
            (datetime.datetime(2021, 3, 4, 23, 59), datetime.time(14, 0)): 1,
            (datetime.datetime(2021, 3, 4, 23, 59), datetime.time(19, 0)): 3,
            (datetime.datetime(2021, 3, 4, 23, 59), datetime.time(22, 0)): 3,
            (datetime.datetime(2021, 3, 5, 0, 0),   datetime.time(10, 0)): 0,
            (datetime.datetime(2021, 3, 5, 0, 0),   datetime.time(14, 0)): 0,
            (datetime.datetime(2021, 3, 5, 0, 0),   datetime.time(19, 0)): 2,
            (datetime.datetime(2021, 3, 5, 0, 0),   datetime.time(22, 0)): 2,
            (datetime.datetime(2021, 3, 5, 12, 15), datetime.time(10, 0)): 3,
            (datetime.datetime(2021, 3, 5, 12, 15), datetime.time(14, 0)): 0,
            (datetime.datetime(2021, 3, 5, 12, 15), datetime.time(19, 0)): 2,
            (datetime.datetime(2021, 3, 5, 12, 15), datetime.time(22, 0)): 2,
            (datetime.datetime(2021, 3, 5, 23, 59), datetime.time(10, 0)): 3,
            (datetime.datetime(2021, 3, 5, 23, 59), datetime.time(14, 0)): 3,
            (datetime.datetime(2021, 3, 5, 23, 59), datetime.time(19, 0)): 2,
            (datetime.datetime(2021, 3, 5, 23, 59), datetime.time(22, 0)): 2,
            (datetime.datetime(2021, 3, 6, 0, 0),   datetime.time(10, 0)): 2,
            (datetime.datetime(2021, 3, 6, 0, 0),   datetime.time(14, 0)): 2,
            (datetime.datetime(2021, 3, 6, 0, 0),   datetime.time(19, 0)): 1,
            (datetime.datetime(2021, 3, 6, 0, 0),   datetime.time(22, 0)): 1,
            (datetime.datetime(2021, 3, 6, 12, 15), datetime.time(10, 0)): 2,
            (datetime.datetime(2021, 3, 6, 12, 15), datetime.time(14, 0)): 2,
            (datetime.datetime(2021, 3, 6, 12, 15), datetime.time(19, 0)): 1,
            (datetime.datetime(2021, 3, 6, 12, 15), datetime.time(22, 0)): 1,
            (datetime.datetime(2021, 3, 6, 23, 59), datetime.time(10, 0)): 2,
            (datetime.datetime(2021, 3, 6, 23, 59), datetime.time(14, 0)): 2,
            (datetime.datetime(2021, 3, 6, 23, 59), datetime.time(19, 0)): 1,
            (datetime.datetime(2021, 3, 6, 23, 59), datetime.time(22, 0)): 1,
            (datetime.datetime(2021, 3, 7, 0, 0),   datetime.time(10, 0)): 1,
            (datetime.datetime(2021, 3, 7, 0, 0),   datetime.time(14, 0)): 1,
            (datetime.datetime(2021, 3, 7, 0, 0),   datetime.time(19, 0)): 0,
            (datetime.datetime(2021, 3, 7, 0, 0),   datetime.time(22, 0)): 0,
            (datetime.datetime(2021, 3, 7, 12, 15), datetime.time(10, 0)): 1,
            (datetime.datetime(2021, 3, 7, 12, 15), datetime.time(14, 0)): 1,
            (datetime.datetime(2021, 3, 7, 12, 15), datetime.time(19, 0)): 0,
            (datetime.datetime(2021, 3, 7, 12, 15), datetime.time(22, 0)): 0,
            (datetime.datetime(2021, 3, 7, 23, 59), datetime.time(10, 0)): 1,
            (datetime.datetime(2021, 3, 7, 23, 59), datetime.time(14, 0)): 1,
            (datetime.datetime(2021, 3, 7, 23, 59), datetime.time(19, 0)): 1,
            (datetime.datetime(2021, 3, 7, 23, 59), datetime.time(22, 0)): 1,
        })

        for (now, tps), delta in samples.items():
            with freezegun.freeze_time(now):
                result = next_occurence(tps)

            self.assertEqual(tps, result.time(), msg=(now, tps))
            self.assertEqual(now.date() + datetime.timedelta(days=delta),
                                 result.date(), msg=(now, tps))



class TestToolsSplittersLoggers(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.tools split/log functions (section 6)."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()


    def test_smooth_split(self):
        """Unit tests for tools.smooth_split function."""
        # def smooth_split(mess, N=1990, sep='\n', rep='')
        smooth_split = tools.smooth_split

        # simple
        self.assertEqual(smooth_split("oui"), ["oui"])
        self.assertEqual(smooth_split("\noui\n"), ["\noui\n"])
        self.assertEqual(smooth_split("o"*1990), ["o"*1990])
        self.assertEqual(smooth_split("o"*1990 + "\n"), ["o"*1990])
        self.assertEqual(smooth_split("o"*1991), ["o"*1990, "o"])
        self.assertEqual(smooth_split("o"*1000 + "\n" + "o"*1000),
                         ["o"*1000, "o"*1000])
        self.assertEqual(smooth_split("o"*1990 + "\n" + "o"*1000),
                         ["o"*1990, "o"*1000])
        self.assertEqual(smooth_split("o"*1991 + "\n" + "o"*1000),
                         ["o"*1990, "o\n" + "o"*1000])
        self.assertEqual(smooth_split("o"*3981), ["o"*1990, "o"*1990, "o"])
        self.assertEqual(smooth_split("o"*2590 + "\n" + "o"*1000),
                         ["o"*1990, "o"*600 + "\n" + "o"*1000])
        self.assertEqual(smooth_split("o"*2590 + "\n" + "o"*1500),
                         ["o"*1990, "o"*600, "o"*1500])

        # custom N
        self.assertEqual(smooth_split("oui", N=30), ["oui"])
        self.assertEqual(smooth_split("\noui\n", N=30), ["\noui\n"])
        self.assertEqual(smooth_split("o"*30, N=30), ["o"*30])
        self.assertEqual(smooth_split("o"*30 + "\n", N=30), ["o"*30])
        self.assertEqual(smooth_split("o"*31, N=30), ["o"*30, "o"])
        self.assertEqual(smooth_split("o"*20 + "\n" + "o"*20, N=30),
                         ["o"*20, "o"*20])
        self.assertEqual(smooth_split("o"*30 + "\n" + "o"*20, N=30),
                         ["o"*30, "o"*20])
        self.assertEqual(smooth_split("o"*31 + "\n" + "o"*20, N=30),
                         ["o"*30, "o\n" + "o"*20])
        self.assertEqual(smooth_split("o"*61, N=30), ["o"*30, "o"*30, "o"])
        self.assertEqual(smooth_split("o"*36 + "\n" + "o"*20, N=30),
                         ["o"*30, "o"*6 + "\n" + "o"*20])
        self.assertEqual(smooth_split("o"*36 + "\n" + "o"*27, N=30),
                         ["o"*30, "o"*6, "o"*27])

        # custom N and sep
        self.assertEqual(smooth_split("oui", N=30, sep="$"), ["oui"])
        self.assertEqual(smooth_split("$oui$", N=30, sep="$"), ["$oui$"])
        self.assertEqual(smooth_split("o"*30, N=30, sep="$"), ["o"*30])
        self.assertEqual(smooth_split("o"*30 + "$", N=30, sep="$"), ["o"*30])
        self.assertEqual(smooth_split("o"*31, N=30, sep="$"), ["o"*30, "o"])
        self.assertEqual(smooth_split("o"*20 + "$" + "o"*20, N=30, sep="$"),
                         ["o"*20, "o"*20])
        self.assertEqual(smooth_split("o"*30 + "$" + "o"*20, N=30, sep="$"),
                         ["o"*30, "o"*20])
        self.assertEqual(smooth_split("o"*31 + "$" + "o"*20, N=30, sep="$"),
                         ["o"*30, "o$" + "o"*20])
        self.assertEqual(smooth_split("o"*61, N=30, sep="$"),
                         ["o"*30, "o"*30, "o"])
        self.assertEqual(smooth_split("o"*36 + "$" + "o"*20, N=30, sep="$"),
                         ["o"*30, "o"*6 + "$" + "o"*20])
        self.assertEqual(smooth_split("o"*36 + "$" + "o"*27, N=30, sep="$"),
                         ["o"*30, "o"*6, "o"*27])

        # custom N and rep
        self.assertEqual(smooth_split("oui", N=30, rep="<REP>"), ["oui"])
        self.assertEqual(smooth_split("\noui\n", N=30, rep="<REP>"),
                         ["\noui\n"])
        self.assertEqual(smooth_split("o"*30, N=30, rep="<REP>"), ["o"*30])
        self.assertEqual(smooth_split("o"*30 + "\n", N=30, rep="<REP>"),
                         ["o"*30 + "<REP>"])
        self.assertEqual(smooth_split("o"*31, N=30, rep="<REP>"),
                         ["o"*30 + "<REP>", "o"])
        self.assertEqual(
            smooth_split("o"*20 + "\n" + "o"*20, N=30, rep="<REP>"),
            ["o"*20 + "<REP>", "o"*20])
        self.assertEqual(
            smooth_split("o"*30 + "\n" + "o"*20, N=30, rep="<REP>"),
            ["o"*30 + "<REP>", "o"*20])
        self.assertEqual(
            smooth_split("o"*31 + "\n" + "o"*20, N=30, rep="<REP>"),
            ["o"*30 + "<REP>", "o\n" + "o"*20])
        self.assertEqual(smooth_split("o"*61, N=30, rep="<REP>"),
                         ["o"*30 + "<REP>", "o"*30 + "<REP>", "o"])
        self.assertEqual(
            smooth_split("o"*36 + "\n" + "o"*20, N=30, rep="<REP>"),
            ["o"*30 + "<REP>", "o"*6 + "\n" + "o"*20])
        self.assertEqual(
            smooth_split("o"*36 + "\n" + "o"*27, N=30, rep="<REP>"),
            ["o"*30 + "<REP>", "o"*6 + "<REP>", "o"*27])


    @mock.patch("lgrez.blocs.tools.smooth_split")
    async def test_send_blocs(self, ss_patch):
        """Unit tests for tools.send_blocs function."""
        # async def send_blocs(messageable, mess, *, N=1990, sep='\n', rep='')
        send_blocs = tools.send_blocs
        # simple
        mgbl, mess = mock.Mock(send=mock.AsyncMock()), mock.Mock()
        ss_patch.return_value = ["oui"]
        messages = await send_blocs(mgbl, mess)
        ss_patch.assert_called_once_with(mess, N=1990, sep='\n', rep='')
        mgbl.send.assert_called_once_with("oui")
        self.assertEqual(messages, [mgbl.send.return_value])
        ss_patch.reset_mock()
        # complex
        mgbl, mess = mock.Mock(send=mock.AsyncMock()), mock.Mock()
        ss_patch.return_value = ["oui", "non", "meh"]
        retmess = [mock.Mock(), mock.Mock(), mock.Mock()]
        mgbl.send.side_effect = retmess
        messages = await send_blocs(mgbl, mess, N=12, sep='$', rep='ùùù')
        ss_patch.assert_called_once_with(mess, N=12, sep='$', rep='ùùù')
        self.assertEqual(mgbl.send.call_count, 3)
        mgbl.send.assert_has_calls([mock.call("oui"), mock.call("non"),
                                    mock.call("meh")])
        self.assertEqual(messages, retmess)
        ss_patch.reset_mock()


    @mock.patch("lgrez.blocs.tools.smooth_split")
    @mock.patch("lgrez.blocs.tools.code_bloc")
    async def test_send_code_blocs(self, cb_patch, ss_patch):
        """Unit tests for tools.send_code_blocs function."""
        # async def send_code_blocs(messageable, mess, *, N=1990, sep='\n',
        #                           rep='', prefixe="", langage="")
        send_code_blocs = tools.send_code_blocs

        # simple
        mgbl, mess = mock.Mock(send=mock.AsyncMock()), "groolk"
        ss_patch.return_value = ["oui"]
        messages = await send_code_blocs(mgbl, mess)
        ss_patch.assert_called_once_with("groolk", N=1990, sep='\n', rep='')
        cb_patch.assert_called_once_with("oui", langage="")
        mgbl.send.assert_called_once_with(cb_patch.return_value)
        self.assertEqual(messages, [mgbl.send.return_value])
        ss_patch.reset_mock()
        cb_patch.reset_mock()

        # complex
        mgbl, mess = mock.Mock(send=mock.AsyncMock()), "groolk"
        ss_patch.return_value = ["oui", "non", "meh"]
        retmess = [mock.Mock(), mock.Mock(), mock.Mock()]
        mgbl.send.side_effect = retmess
        retcbs = [mock.Mock(), mock.Mock(), mock.Mock()]
        cb_patch.side_effect = retcbs
        messages = await send_code_blocs(mgbl, mess, N=12, sep='$', rep='ùùù',
                                         langage="oh")
        ss_patch.assert_called_once_with("groolk", N=12, sep='$', rep='ùùù')
        self.assertEqual(cb_patch.call_count, 3)
        cb_patch.assert_has_calls([mock.call("oui", langage="oh"),
                                   mock.call("non", langage="oh"),
                                   mock.call("meh", langage="oh")])
        self.assertEqual(mgbl.send.call_count, 3)
        mgbl.send.assert_has_calls([mock.call(mk) for mk in retcbs])
        self.assertEqual(messages, retmess)
        ss_patch.reset_mock()
        cb_patch.reset_mock()

        # with prefix
        mgbl, mess = mock.Mock(send=mock.AsyncMock()), "groolk"
        ss_patch.return_value = ["PREF::\noui", "non", "meh"]
        retmess = [mock.Mock(), mock.Mock(), mock.Mock()]
        mgbl.send.side_effect = retmess
        retcbs = ["<<oui>>", "<<non>>", "<<meh>>"]
        cb_patch.side_effect = retcbs
        messages = await send_code_blocs(mgbl, mess, N=12, sep='$', rep='ùùù',
                                         langage="oh", prefixe="PREF::")
        ss_patch.assert_called_once_with("PREF::\ngroolk", N=12, sep='$',
                                         rep='ùùù')
        self.assertEqual(cb_patch.call_count, 3)
        cb_patch.assert_has_calls([mock.call("oui", langage="oh"),
                                   mock.call("non", langage="oh"),
                                   mock.call("meh", langage="oh")])
        self.assertEqual(mgbl.send.call_count, 3)
        mgbl.send.assert_has_calls([mock.call("PREF::\n<<oui>>"),
                                    mock.call("<<non>>"),
                                    mock.call("<<meh>>")])
        self.assertEqual(messages, retmess)
        ss_patch.reset_mock()
        cb_patch.reset_mock()


    @mock.patch("lgrez.blocs.tools.send_blocs")
    @mock.patch("lgrez.blocs.tools.send_code_blocs")
    async def test_log(self, scb_patch, sb_patch):
        """Unit tests for tools.log function."""
        # async def log(message, *, code=False, N=1990, sep='\n', rep='',
        #               prefixe="", langage="")
        log = tools.log

        # simple - code False
        messages = await log("groolk")
        sb_patch.assert_called_once_with(
            config.Channel.logs, "groolk", N=1990, sep='\n', rep='')
        self.assertEqual(messages, sb_patch.return_value)
        sb_patch.reset_mock()
        scb_patch.assert_not_called()

        # simple - code True
        messages = await log("groolk", code=True)
        scb_patch.assert_called_once_with(
            config.Channel.logs, "groolk", N=1990, sep='\n', rep='',
            prefixe="", langage="")
        self.assertEqual(messages, scb_patch.return_value)
        scb_patch.reset_mock()
        sb_patch.assert_not_called()

        # complex - code False
        messages = await log("groolk", N=12, sep='$', rep='ùùù',
                             langage="oh", prefixe="PREF::")
        sb_patch.assert_called_once_with(
            config.Channel.logs, "PREF::\ngroolk", N=12, sep='$', rep='ùùù')
        self.assertEqual(messages, sb_patch.return_value)
        sb_patch.reset_mock()
        scb_patch.assert_not_called()

        # complex - code True
        messages = await log("groolk", code=True, N=12, sep='$', rep='ùùù',
                             langage="oh", prefixe="PREF::")
        scb_patch.assert_called_once_with(
            config.Channel.logs, "groolk", N=12, sep='$', rep='ùùù',
            prefixe="PREF::", langage="oh")
        self.assertEqual(messages, scb_patch.return_value)
        scb_patch.reset_mock()
        sb_patch.assert_not_called()



class TestToolsMiscellaneous(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.tools misc. functions (sections 7 & 8)."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()


    @mock.patch("lgrez.bdd.Joueur.from_member")
    async def test_create_context(self, jfm_patch):
        """Unit tests for tools.create_context function."""
        # async def create_context(member, content)
        create_context = tools.create_context
        member = mock.Mock()
        mess = mock.Mock()
        chan = jfm_patch.return_value.private_chan
        chan.history.return_value.flatten = mock.AsyncMock(return_value=[mess])

        ctx = await create_context(member, "ouiZZ")
        chan.history.assert_called_once_with(limit=1)
        jfm_patch.assert_called_once_with(member)
        config.bot.get_context.assert_called_once_with(mock.ANY)
        self.assertEqual(mess, config.bot.get_context.call_args.args[0])
        self.assertEqual(mess.author, member)
        self.assertEqual(mess.content, "ouiZZ")
        self.assertEqual(ctx, config.bot.get_context.return_value)


    def test_remove_accents(self):
        """Unit tests for tools.remove_accents function."""
        # def remove_accents(text)
        remove_accents = tools.remove_accents
        samples = {
            "": "",
            "allo": "allo",
            "AllO": "AllO",
            "énorme ÉNORME": "enorme ENORME",
            "àéïìùÀÉÏÌÙ‰": "aeiiuAEIIU‰",
            "énorme?♥♦♣♠": "enorme?♥♦♣♠",
        }
        for sample, result in samples.items():
            self.assertEqual(remove_accents(sample), result)

    @unittest.SkipTest
    def test_eval_accols(rep, globals_=None, locals_=None, debug=False):
        """Unit tests for tools.eval_accols function."""
        # def eval_accols(text)
        eval_accols = tools.eval_accols

    def test_bold(self):
        """Unit tests for tools.bold function."""
        # def bold(text)
        bold = tools.bold
        self.assertEqual(bold("sample"), "**sample**")

    def test_ital(self):
        """Unit tests for tools.ital function."""
        # def ital(text)
        ital = tools.ital
        self.assertEqual(ital("sample"), "*sample*")

    def test_soul(self):
        """Unit tests for tools.soul function."""
        # def soul(text)
        soul = tools.soul
        self.assertEqual(soul("sample"), "__sample__")

    def test_strike(self):
        """Unit tests for tools.strike function."""
        # def strike(text)
        strike = tools.strike
        self.assertEqual(strike("sample"), "~~sample~~")

    def test_code(self):
        """Unit tests for tools.code function."""
        # def code(text)
        code = tools.code
        self.assertEqual(code("sample"), "`sample`")

    def test_code_bloc(self):
        """Unit tests for tools.code_bloc function."""
        # def code_bloc(text, language="")
        code_bloc = tools.code_bloc
        self.assertEqual(code_bloc("sample"), "```\nsample```")
        self.assertEqual(code_bloc("sample", langage="bzz"),
                         "```bzz\nsample```")

    def test_quote(self):
        """Unit tests for tools.quote function."""
        # def quote(text)
        quote = tools.quote
        self.assertEqual(quote("sample"), "> sample")

    def test_quote_bloc(self):
        """Unit tests for tools.quote_bloc function."""
        # def quote_bloc(text)
        quote_bloc = tools.quote_bloc
        self.assertEqual(quote_bloc("sample"), ">>> sample")

    def test_spoiler(self):
        """Unit tests for tools.spoiler function."""
        # def spoiler(text)
        spoiler = tools.spoiler
        self.assertEqual(spoiler("sample"), "||sample||")



import asyncio
import functools
import os


class _PatchEnv():
    def __init__(self, **kwargs):
        self.vars = kwargs

    def __enter__(self):
        self.olds = {var: os.getenv(var) for var in self.vars}
        for var, value in self.vars.items():
            if value is None:
                del os.environ[var]
            else:
                os.environ[var] = value
        return self

    def __exit__(self, etype, eval, tb):
        for var, value in self.olds.items():
            if value is None:
                del os.environ[var]
            else:
                os.environ[var] = value

    def __call__(self, func):
        if asyncio.iscoroutinefunction(func):
            @functools.wraps(func)
            async def newfunc(*args, **kwargs):
                self.__enter__()
                try:
                    await func(*args, **kwargs)
                finally:
                    self.__exit__(None, None, None)
        else:
            @functools.wraps(func)
            def newfunc(*args, **kwargs):
                self.__enter__()
                try:
                    func(*args, **kwargs)
                finally:
                    self.__exit__(None, None, None)

        return newfunc


def patch_env(**kwargs):
    """Patch environment variables.

    Args:
        kwargs: NAME, value pair(s) of environment variables to patch,
            restored (or deleted) after. Set a variable to ``None`` to
            temporary delete it.

    Can be used as a function decorator or context manager, as for
    `unittest.mock.patch`.
    """
    return _PatchEnv(**kwargs)



"""lg-rez / Gestion des données

Déclaration de toutes les tables, colonnes, méthodes, et connection à la BDD

"""

import sqlalchemy
import psycopg2

from lgrez.bdd import base
from lgrez.bdd.enums import *
from lgrez.bdd.model_joueurs import *
from lgrez.bdd.model_jeu import *
from lgrez.bdd.model_actions import *
from lgrez.bdd.model_ia import *
# All tables and enums directly accessible via bdd.<name>


#: __all__ = toutes les classes de données publiques
__all__ = [nom for nom in base.tables if not nom.startswith("_")]


tables = base.tables

connect = base.connect

SQLAlchemyError = sqlalchemy.exc.SQLAlchemyError

DriverOperationalError = psycopg2.OperationalError



"""lg-rez / features / Communication

Envoi de messages, d'embeds...

"""
import datetime
import functools
import os
import re

import discord
from discord.ext import commands
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from lgrez import config
from lgrez.blocs import tools, gsheets
from lgrez.bdd import (Joueur, Action, Camp, BaseAction, Utilisation,
                       Statut, ActionTrigger, CandidHaroType, UtilEtat, Vote)
from lgrez.features import gestion_actions
from lgrez.features.sync import transtype


def _joueur_repl(mtch):
    """Remplace @(Prénom Nom) par la mention du joueur, si possible"""
    nearest = Joueur.find_nearest(mtch.group(1),
                                  col=Joueur.nom, sensi=0.8)
    if nearest:
        joueur = nearest[0][0]
        try:
            return joueur.member.mention
        except ValueError:
            pass
    return mtch.group(1)

def _role_repl(mtch):
    """Remplace @role_slug par la mention du rôle"""
    if mtch.group(1).lower() in config.Role:
        role = getattr(config.Role, mtch.group(1))
        return role.mention
    return mtch.group(1)

def _emoji_repl(mtch):
    """Remplace :(emoji): par la représentation de l'emoji, si possible"""
    emo = tools.emoji(mtch.group(1), must_be_found=False)
    if emo:
        return str(emo)
    return mtch.group()


async def _create_annoncemort_embed(ctx, victime=None):
    joueur = await tools.boucle_query_joueur(ctx, victime,
                                             "Qui est la victime ?")

    role = joueur.role.nom_complet
    mess = await ctx.send(
        f"Rôle à afficher pour {joueur.nom} = {role} ? "
        "(Pas double peau ou autre)"
    )
    if await tools.yes_no(mess):
        emoji_camp = joueur.camp.discord_emoji_or_none
    else:
        await ctx.send("Rôle à afficher :")
        role = (await tools.wait_for_message_here(ctx)).content
        mess = await ctx.send("Camp :")
        camps = Camp.query.filter_by(public=True).all()
        emoji_camp = await tools.wait_for_react_clic(
            mess,
            [camp.discord_emoji for camp in camps if camp.emoji]
        )

    if joueur.statut == Statut.MV:
        mess = await ctx.send("Annoncer la mort-vivance ?")
        if await tools.yes_no(mess):
            role += " Mort-Vivant"
        else:
            emoji_camp = joueur.role.camp.discord_emoji_or_none

    await ctx.send("Contexte ?")
    desc = (await tools.wait_for_message_here(ctx)).content

    # Création embed
    embed = discord.Embed(
        title=f"Mort de {tools.bold(joueur.nom)}, {role}",
        description=desc,
        color=0x730000
    )
    embed.set_author(name="Oh mon dieu, quelqu'un est mort !")
    if emoji_camp:
        embed.set_thumbnail(url=emoji_camp.url)

    return embed


class Communication(commands.Cog):
    """Commandes d'envoi de messages, d'embeds, d'annonces..."""

    current_embed = None

    @commands.command()
    @tools.mjs_only
    async def embed(self, ctx, key=None, *, val=None):
        """Prépare un embed (message riche) et l'envoie (COMMANDE MJ)

        Warning:
            Commande en bêta, non couverte par les tests unitaires
            et souffrant de bugs connus (avec les fields notemment)

        Args:
            key: sous-commande (voir ci-dessous). Si omis, prévisualise
                le brouillon d'embed actuellement en préparation ;
            val: valeur associée. Pour les sous-commandes de
                construction d'élement, supprime ledit élément si omis.

        - Sous-commandes générales :
            - ``!embed create <titre>`` :  Créer un nouveau brouillon
              d'embed (un seul brouillon en parallèle, partout)
            - ``!embed delete`` :          Supprimer le brouillon d'embed
            - ``!embed preview``  :        Voir l'embed sans les aides
            - ``!embed post [#channel]`` : Envoyer l'embed sur ``#channel``
              (chan courant si omis)

        - Sous-commandes de construction d'éléments :
            - Éléments généraux :
                - ``!embed title [titre]``
                - ``!embed description [texte]``
                - ``!embed url [url*]``
                - ``!embed color [#ffffff]`` (barre de gauche,
                  code hexadécimal)

            - Auteur :
                - ``!embed author [nom]``
                - ``!embed author_url [url*]``
                - ``!embed author_icon [url**]``

            - Talon :
                - ``!embed footer [texte]``
                - ``!embed footer_icon [url**]``

            - Images :
                - ``!embed image [url**]`` (grande image)
                - ``!embed thumb [url**]`` (en haut à droite)

            - Champs : syntaxe spéciale
                - ``!embed field <i> <skey> [val]``
                    - ``i`` : Numéro du champ (commençant à ``0``).
                      Si premier champ non existant, le crée ;
                    - ``skey`` :
                        - ``name`` :    Nom du champ
                        - ``value`` :   Valeur du champ
                        - ``delete`` :  Supprime le champ

                Les champs sont (pour l'instant) forcément de type
                inline (côte à côte).

        \* Les URL doivent commencer par http(s):// pour être
        reconnues comme telles.

        \*\* Ces URL doivent correspondre à une image.
        """

        if val is None:
            val = discord.Embed.Empty

        # Récupération de l'embed (stocké dans le cog)
        emb = self.current_embed

        # Attributs modifiables directement : emb.attr = value
        direct = [
            "title",
            "description",
            "url",
        ]
        # Attributs à modifier en appelant une méthode : emb.set_<attr>(value)
        # avec method[key] = (<attr>, <value>)
        method = {
            "footer": ("footer", "text"),
            "footer_icon": ("footer", "icon_url"),
            "image": ("image", "url"),
            "thumb": ("thumbnail", "url"),
            "author_url": ("author", "url"),
            "author_icon": ("author", "icon_url"),
        }

        if not emb:             # Pas d'embed en cours
            if key == "create" and val:
                emb = discord.Embed(title=val)
            else:
                await ctx.send(
                    "Pas d'embed en préparation. "
                    + tools.code("!embed create <titre>")
                    + " pour en créer un.")
                return

        elif key in direct:         # Attributs modifiables directement
            setattr(emb, key, val)

        elif key in method:         # Attributs à modifier via une méthode
            prop, attr = method[key]
            getattr(emb, f"set_{prop}")(**{attr: val})

        elif key == "author":       # Cas particulier
            emb.set_author(name=val) if val else emb.remove_author()

        elif key == "color":        # Cas particulier : cast couleur en int
            try:
                if val:
                    emb.color = eval(val.replace("#", "0x"))
                else:
                    emb.color = discord.Embed.Empty
            except Exception:
                await ctx.send("Couleur invalide")
                return

        elif key == "field":        # Cas encore plus particulier
            i_max = len(emb.fields)     # N fields ==> i_max = N+1
            try:
                i, skey, val = val.split(" ", maxsplit=2)
                i = int(i)
                if i < 0 or i > i_max:
                    await ctx.send("Numéro de field invalide")
                    return
                if skey not in ["name", "value", "delete"]:
                    raise ValueError()
            except Exception:
                await ctx.send("Syntaxe invalide")
                return

            if i == i_max:
                if skey == "name":
                    emb.add_field(name=val,
                                  value=f"!embed field {i} value <valeur>")
                elif skey == "value":
                    emb.add_field(name=f"!embed field {i} name <nom>",
                                  value=val)
                # emb.add_field(*, name, value, inline=True)

            else:
                if skey == "name":
                    emb.set_field_at(i, name=val, value=emb.fields[i].value)
                elif skey == "value":
                    emb.set_field_at(i, name=emb.fields[i].name, value=val)
                else:
                    emb.remove_field(i)
                # emb.set_field_at(i, *, name, value, inline=True)

        elif key == "delete":
            self.current_embed = None
            await ctx.send(
                "Supprimé. " + tools.code("!embed create <titre>")
                + " pour en créer un."
            )
            return

        elif key == "create":
            await ctx.send(
                "Déjà un embed en cours de création. Utiliser "
                + tools.code("!embed delete") + "pour le supprimer."
            )

        elif key == "preview":
            await ctx.send("Prévisuatisation :", embed=emb)
            await ctx.send(
                "Utiliser " + tools.code("!embed post #channel")
                + "pour publier l'embed."
            )
            return

        elif key == "post":
            if not val:     # channel non précisé
                await ctx.send(embed=emb)
            elif (chan := tools.channel(val, must_be_found=False)):
                await chan.send(embed=emb)
                await ctx.send("Et pouf !")
            else:
                await ctx.send(
                    f"Channel inconnu. Réessaye en le mentionnant "
                    f"({tools.code('#channel')})"
                )
            return

        elif key is not None:
            await ctx.send(
                f"Option {key} incorrecte ; voir "
                + tools.code("!help embed") + "pour en savoir plus."
            )
            return

        h_emb = emb.copy()
        if not emb.title:
            h_emb.title = "!embed title <titre>"
        if not emb.description:
            h_emb.description = "!embed description <description>"
        if not emb.footer:
            h_emb.set_footer(text="!embed footer <footer>")
        if not emb.author:
            h_emb.set_author(name="!embed author <auteur>")

        i_max = len(emb.fields)                 # N fields ==> i_max = N+1
        h_emb.add_field(name=f"!embed field {i_max} name <nom>",
                        value=f"!embed field {i_max} value <nom>")

        await ctx.send("Embed en préparation :", embed=h_emb)
        await ctx.send(
            f"Utiliser {tools.code('!embed preview')} pour prévisualiser "
            f"l'embed.\n Autres options : "
            + tools.code("!embed color <#xxxxxx> / url <url> / image <url> / "
                         "thumb <url> / author_url <url> / footer_icon <url>")
        )

        self.current_embed = emb


    @commands.command(aliases=["tell"])
    @tools.mjs_only
    async def send(self, ctx, cible, *, message):
        """Envoie un message à tous ou certains joueurs (COMMANDE MJ)

        Args:
            cible: destinataires
            message: message, éventuellement formaté

        ``cible`` peut être :
            - ``all`` :             Tous les joueurs inscrits, vivants et morts
            - ``vivants`` :         Les joueurs en vie
            - ``morts`` :           Les joueurs morts
            - ``<crit>=<filtre>`` : Les joueurs répondant au critère
              ``Joueur.<crit> == <filtre>``. ``crit`` peut être ``"nom"``,
              ``"chambre"``, ``"statut"``, ``"role"``, ``"camp"``...
              L'ensemble doit être entouré de guillements si ``filtre``
              contient un espace. Les rôles/camps sont cherchés par slug.
            - *le nom d'un joueur*  (raccourci pour ``nom=X``, doit être
              entouré de guillements si nom + prénom)

        ``message`` peut contenir un ou plusieurs bouts de code Python
        à évaluer, entourés d'accolades.

        L'évaluation est faite séparément pour chaque joueur, ce qui
        permet de personnaliser le message grâce aux variables
        particulières dépendant du joueur :

            - ``joueur`` :  objet BDD du joueur recevant le message
              ==> ``joueur.nom``, ``joueur.role``...
            - ``member`` :  objet :class:`discord.Member` associé
              ==> ``member.mention``
            - ``chan`` :    objet :class:`discord.TextChannel` du
              chan privé du joueur

        Attention :
            - ``ctx`` :     objet :class:`discord.ext.commands.Context`
              de ``!send``  ==> ``ctx.author`` = lanceur de la commande !!!

        Les différentes tables de données sont accessibles sous leur nom
        (``Joueur``, ``Role``...)

        Il est impossible d'appeller des coroutines (await) dans le code
        à évaluer.

        Examples:
            - ``!send all Bonsoir à tous c'est Fanta``
            - ``!send vivants Attention {member.mention},
              derrière toi c'est affreux !``
            - ``!send "role=servante" Ça va vous ?
              Vous êtes bien {joueur.role.nom} ?``
        """
        if cible == "all":
            joueurs = Joueur.query.all()
        elif cible == "vivants":
            joueurs = Joueur.query.filter(Joueur.est_vivant).all()
        elif cible == "morts":
            joueurs = Joueur.query.filter(Joueur.est_mort).all()
        elif "=" in cible:
            crit, _, filtre = cible.partition("=")
            crit = crit.strip()
            if crit in Joueur.attrs:
                col = Joueur.attrs[crit]
                arg = transtype(filtre.strip(), col)
                joueurs = Joueur.query.filter_by(**{crit: arg}).all()
            else:
                raise commands.UserInputError(f"critère '{crit}' incorrect")
        else:
            joueurs = [await tools.boucle_query_joueur(ctx, cible, "À qui ?")]

        if not joueurs:
            await ctx.send("Aucun joueur trouvé.")
            return

        await ctx.send(f"{len(joueurs)} trouvé(s), envoi...")
        for joueur in joueurs:
            member = joueur.member
            chan = joueur.private_chan

            evaluated_message = tools.eval_accols(message, locals_=locals())
            await chan.send(evaluated_message)

        await ctx.send("Fini.")


    @commands.command()
    @tools.mjs_only
    async def post(self, ctx, chan, *, message):
        """Envoie un message dans un salon (COMMANDE MJ)

        Args:
            chan: nom du salon ou sa mention
            message: message à envoyer (peut être aussi long que
                nécessaire, contenir des sauts de lignes...)
        """
        chan = tools.channel(chan)
        await chan.send(message)
        await ctx.send("Fait.")


    @commands.command()
    @tools.mjs_only
    async def plot(self, ctx, quoi, depuis=None):
        """Trace le résultat du vote et l'envoie sur #annonces (COMMANDE MJ)

        Warning:
            Commande en bêta, non couverte par les tests unitaires

        Args:
            quoi: peut être

                - ``cond``   pour le vote pour le condamné
                - ``maire``  pour l'élection à la Mairie

            depuis: heure éventuelle à partir de laquelle compter les
                votes (si plusieurs votes dans la journée), compte tous
                les votes du jour par défaut. Si plus tard que l'heure
                actuelle, compte les votes de la veille.

        Trace les votes sous forme d'histogramme à partir du Tableau de
        bord, en fait un embed en présisant les résultats détaillés et
        l'envoie sur le chan ``#annonces``.

        Si ``quoi == "cond"``, déclenche aussi les actions liées au mot
        des MJs (:attr:`.bdd.ActionTrigger.mot_mjs`).
        """
        # Différences plot cond / maire
        if quoi == "cond":
            vote_enum = Vote.cond
            haro_candidature = CandidHaroType.haro
            typo = "bûcher du jour"
            mort_election = "Mort"
            pour_contre = "contre"
            emoji = config.Emoji.bucher
            couleur = 0x730000

        elif quoi == "maire":
            vote_enum = Vote.maire
            haro_candidature = CandidHaroType.candidature
            typo = "nouveau maire"
            mort_election = "Élection"
            pour_contre = "pour"
            emoji = config.Emoji.maire
            couleur = 0xd4af37

        else:
            raise commands.BadArgument("`quoi` doit être `maire` ou `cond`")

        if depuis:
            tps = tools.heure_to_time(depuis)
        else:
            tps = datetime.time(0, 0)

        ts = datetime.datetime.combine(datetime.date.today(), tps)
        if ts > datetime.datetime.now():        # hier
            ts -= datetime.timedelta(days=1)

        log = f"!plot {quoi} (> {ts}) :"
        query = Utilisation.query.filter(
            Utilisation.etat == UtilEtat.validee,
            Utilisation.ts_decision > ts,
            Utilisation.action.has(active=True),
        )
        cibles = {}

        # Get votes
        utils = query.filter(Utilisation.action.has(vote=vote_enum)).all()
        votes = {util.action.joueur: util.cible for util in utils}
        votelog = " / ".join(f'{v.nom} -> {c.nom}' for v, c in votes.items())
        log += f"\n  - Votes : {votelog}"

        for votant, vote in votes.items():
            cibles.setdefault(vote, [])
            cibles[vote].append(votant.nom)

        # Get intriguants
        intba = BaseAction.query.get(config.modif_vote_baseaction)
        if intba:
            log += "\n  - Intrigant(s) : "
            for util in query.filter(Utilisation.action.has(base=intba)).all():

                votant = util.ciblage("cible").valeur
                vote = util.ciblage("vote").valeur
                log += (f"{util.action.joueur.nom} : "
                        f"{votant.nom} -> {vote.nom} / ")

                initial_vote = votes.get(votant)
                if initial_vote:
                    cibles[initial_vote].remove(votant.nom)
                    if not cibles[initial_vote]:    # plus de votes
                        del cibles[initial_vote]
                votes[votant] = vote
                cibles.setdefault(vote, [])
                cibles[vote].append(votant.nom)

        # Tri des votants
        for votants in cibles.values():
            votants.sort()      # ordre alphabétique

        # Get corbeaux, après tri -> à la fin
        corba = BaseAction.query.get(config.ajout_vote_baseaction)
        if corba:
            log += "\n  - Corbeau(x) : "
            for util in query.filter(Utilisation.action.has(base=corba)).all():
                log += f"{util.action.joueur.nom} -> {util.cible} / "
                cibles.setdefault(util.cible, [])
                cibles[util.cible].extend(
                    [util.action.joueur.role.nom]*config.n_ajouts_votes
                )

        # Classe utilitaire
        @functools.total_ordering
        class _Cible():
            """Représente un joueur ciblé, pour usage dans !plot"""
            def __init__(self, joueur, votants):
                self.joueur = joueur
                self.votants = votants

            def __repr__(self):
                return f"{self.joueur.nom} ({self.votes})"

            def __eq__(self, other):
                if not isinstance(other, type(self)):
                    return NotImplemented
                return (self.joueur.nom == other.joueur.nom
                        and self.votes == other.votes)

            def __lt__(self, other):
                if not isinstance(other, type(self)):
                    return NotImplemented
                if self.votes == other.votes:
                    return (self.joueur.nom < other.joueur.nom)
                return (self.votes < other.votes)

            @property
            def votes(self):
                return len(self.votants)

            @property
            def eligible(self):
                return any(ch.type == haro_candidature
                           for ch in self.joueur.candidharos)

            def couleur(self, choisi):
                if self == choisi:
                    return hex(couleur).replace("0x", "#")
                if self.eligible:
                    return "#64b9e9"
                else:
                    return "gray"

        # Récupération votes
        cibles = [_Cible(jr, vts) for (jr, vts) in cibles.items()]
        cibles.sort(reverse=True)       # par nb de votes, puis ordre alpha
        log += f"\n  - Cibles : {cibles}"

        # Détermination cible
        choisi = None
        eligibles = [c for c in cibles if c.eligible]
        log += f"\n  - Éligibles : {eligibles}"

        if eligibles:
            maxvotes = eligibles[0].votes
            egalites = [c for c in eligibles if c.votes == maxvotes]

            if len(egalites) > 1:       # Égalité
                mess = await ctx.send(
                    "Égalité entre\n"
                    + "\n".join(f"{tools.emoji_chiffre(i+1)} {c.joueur.nom}"
                                for i, c in enumerate(egalites))
                    + "\nQui meurt / est élu ? (regarder vote du maire, "
                    "0️⃣ pour personne / si le vainqueur est garde-loupé, "
                    "inéligible ou autre)"
                )
                choice = await tools.choice(mess, len(egalites), start=0)
                if choice:      # pas 0
                    choisi = eligibles[choice - 1]

            else:
                mess = await ctx.send(
                    "Joueur éligible le plus voté : "
                    + tools.bold(eligibles[0].joueur.nom)
                    + " \nÇa meurt / est élu ? (pas garde-loupé, "
                    "inéligible ou autre)"
                )
                if await tools.yes_no(mess):
                    choisi = eligibles[0]

        log += f"\n  - Choisi : {choisi or '[aucun]'}"
        await tools.log(log)

        # Paramètres plot
        discord_gray = '#2F3136'
        plt.figure(facecolor=discord_gray)
        plt.rcParams.update({'font.size': 16})
        ax = plt.axes(facecolor='#8F9194')  # coloration de TOUT le graphe
        ax.tick_params(axis='both', colors='white')
        ax.spines['bottom'].set_color('white')
        ax.spines['left'].set_color(discord_gray)
        ax.spines['right'].set_color(discord_gray)
        ax.spines['top'].set_color(discord_gray)
        ax.set_facecolor(discord_gray)
        ax.set_axisbelow(True)
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))

        # Plot
        ax.bar(
            x=range(len(cibles)),
            height=[c.votes for c in cibles],
            tick_label=[c.joueur.nom.replace(" ", "\n", 1) for c in cibles],
            color=[c.couleur(choisi) for c in cibles],
        )
        plt.grid(axis="y")

        if not os.path.isdir("figures"):
            os.mkdir("figures")

        now = datetime.datetime.now().strftime("%Y-%m-%d--%H")
        image_path = f"figures/hist_{now}_{quoi}.png"
        plt.savefig(image_path, bbox_inches="tight")

        # --------------- Partie Discord ---------------

        # Détermination rôle et camp
        emoji_camp = None
        if choisi:
            if quoi == "cond":
                role = choisi.joueur.role.nom_complet
                mess = await ctx.send(
                    f"Rôle à afficher pour {choisi.joueur.nom} = {role} ? "
                    "(Pas double peau ou autre)"
                )
                if await tools.yes_no(mess):
                    emoji_camp = choisi.joueur.camp.discord_emoji_or_none
                else:
                    await ctx.send("Rôle à afficher :")
                    role = (await tools.wait_for_message_here(ctx)).content
                    mess = await ctx.send("Camp :")
                    camps = Camp.query.filter_by(public=True).all()
                    emoji_camp = await tools.wait_for_react_clic(
                        mess,
                        [camp.discord_emoji for camp in camps if camp.emoji]
                    )

                nometrole = f"{tools.bold(choisi.joueur.nom)}, {role}"
            else:
                # Maire : ne pas annoncer le rôle
                nometrole = f"{tools.bold(choisi.joueur.nom)}"
        else:
            nometrole = "personne, bande de tocards"

        # Création embed
        embed = discord.Embed(
            title=f"{mort_election} de {nometrole}",
            description=f"{len(votes)} votes au total",
            color=couleur
        )
        embed.set_author(name=f"Résultats du vote pour le {typo}",
                         icon_url=emoji.url)

        if emoji_camp:
            embed.set_thumbnail(url=emoji_camp.url)

        embed.set_footer(text="\n".join(
            ("A" if cible.votes == 1 else "Ont")
            + f" voté {pour_contre} {cible.joueur.nom} : "
            + ", ".join(cible.votants)
            for cible in cibles
        ))

        file = discord.File(image_path, filename="image.png")
        embed.set_image(url="attachment://image.png")

        # Envoi
        mess = await ctx.send("Ça part ?\n", file=file, embed=embed)
        if await tools.yes_no(mess):
            # Envoi du graphe
            file = discord.File(image_path, filename="image.png")
            # Un objet File ne peut servir qu'une fois, il faut le recréer

            await config.Channel.annonces.send(
                "@everyone Résultat du vote ! :fire:",
                file=file,
                embed=embed,
            )
            await ctx.send(
                f"Et c'est parti dans {config.Channel.annonces.mention} !"
            )

            if quoi == "cond":
                # Actions au mot des MJs
                for action in Action.query.filter(Action.base.has(
                        trigger_debut=ActionTrigger.mot_mjs)).all():
                    await gestion_actions.open_action(action)

                await ctx.send("(actions liées au mot MJ ouvertes)")

        else:
            await ctx.send("Mission aborted.")
            self.current_embed = embed


    @commands.command()
    @tools.mjs_only
    async def annoncemort(self, ctx, *, victime=None):
        """Annonce un ou plusieur mort(s) hors-vote (COMMANDE MJ)

        Args:
            victime: (premier) mort à annoncer

        Envoie un embed par mort dans ``#annonces``
        """
        embeds = []
        ok = False
        while not ok:
            embed = await _create_annoncemort_embed(ctx, victime)
            embeds.append(embed)
            victime = None      # que le 1er appel

            mess = await ctx.send("Ajouter un mort ?", embed=embed)
            if not await tools.yes_no(mess):
                ok = True

        mess = await ctx.send("Ça part ?")
        if await tools.yes_no(mess):
            await config.Channel.annonces.send(
                "@everyone Il s'est passé quelque chose ! :scream:",
                embed=embeds[0]
            )
            for embed in embeds[1:]:
                await config.Channel.annonces.send(embed=embed)
            await ctx.send(
                f"Et c'est parti dans {config.Channel.annonces.mention} !"
            )

        else:
            await ctx.send("Mission aborted.")



    @commands.command()
    @tools.mjs_only
    async def lore(self, ctx, doc_id):
        """Récupère et poste un lore depuis un Google Docs (COMMANDE MJ)

        Convertit les formats et éléments suivants vers Discord :
            - Gras, italique, souligné, barré;
            - Petites majuscules (-> majuscules);
            - Polices à chasse fixe (Consolas / Courier New) (-> code);
            - Liens hypertextes;
            - Listes à puces;
            - Mentions de joueurs, sous la forme ``@Prénom Nom``;
            - Emojis, sous la forme ``:nom:``.

        Permet soit de poster directement dans #annonces, soit de
        récupérer la version formatée du texte (pour copier-coller).

        Args:
            doc_id: ID ou URL du document (doit être public ou dans le
                Drive partagé avec le compte de service)
        """
        if len(doc_id) < 44:
            raise commands.BadArgument("'doc_id' doit être l'ID ou l'URL "
                                       "d'un document Google Docs")
        elif len(doc_id) > 44:      # URL fournie (pas que l'ID)
            mtch = re.search(r"/d/([\w-]{44})(\W|$)", doc_id)
            if mtch:
                doc_id = mtch.group(1)
            else:
                raise commands.BadArgument("'doc_id' doit être l'ID ou l'URL "
                                           "d'un document Google Docs")

        await ctx.send("Récupération du document...")
        async with ctx.typing():
            content = gsheets.get_doc_content(doc_id)

        formatted_text = ""
        for (_text, style) in content:
            _text = _text.replace("\v", "\n").replace("\f", "\n")

            # Espaces/newlines au début/fin de _text ==> à part
            motif = re.fullmatch(r"(\s*)(.*?)(\s*)", _text)
            pref, text, suff = motif.group(1, 2, 3)

            if not text:    # espaces/newlines uniquement
                formatted_text += pref + suff
                continue

            # Remplacement des mentions
            text = re.sub(r"@([\w-]+ [\w-]+)", _joueur_repl, text)
            text = re.sub(r"@(\w+)", _role_repl, text)
            text = re.sub(r":(\w+):", _emoji_repl, text)

            if style.get("bold"):
                text = tools.bold(text)
            if style.get("italic"):
                text = tools.ital(text)
            if style.get("strikethrough"):
                text = tools.strike(text)
            if style.get("smallCaps"):
                text = text.upper()
            if (wff := style.get("weightedFontFamily")):
                if wff["fontFamily"] in ["Consolas", "Courier New"] :
                    text = tools.code(text)
            if (link := style.get("link")):
                if (url := link.get("url")):
                    if "://" not in text:
                        text = text + f" (<{url}>)"
            elif style.get("underline"):    # ne pas souligner si lien
                text = tools.soul(text)

            formatted_text += pref + text + suff

        await ctx.send("————————————————————")
        await tools.send_blocs(ctx, formatted_text)
        mess = await ctx.send("————————————————————\nPublier sur "
                              f"{config.Channel.annonces.mention} / "
                              "Récupérer la version formatée / Stop ?")
        r = await tools.wait_for_react_clic(mess, {"📣": 1, "📝": 2, "⏹": 0})

        if r == 1:
            await tools.send_blocs(config.Channel.annonces, formatted_text)
            await ctx.send("Fait !")
        elif r == 2:
            await tools.send_code_blocs(ctx, formatted_text)
        else:
            await ctx.send("Mission aborted.")


    @commands.command()
    @tools.mjs_only
    async def modif(self, ctx, ref, *, modifs):
        """Modifie un message du bot (COMMANDE MJ)

        Args:
            ref: référence vers le message à modifier, peut être

                - le lien du message,
                - son couple d'IDs dans le serveur (salon-message),
                - son ID seul (seulement si message dans ce salon),
                - un nom/mention de salon (si le dernier message est du bot),
                - rien, si on répond au message à modifier.

            modifs: modifications à faire, sous la forme "avant > après"
                (ou juste "> après" pour tout remplacer).

        Il n'est pas possible (pour le moment ?) de modifier une image,
        pièce jointe ou embed.
        """
        # Détermination message à modifier
        if ctx.message.reference:
            msg = ctx.message.reference.resolved
            if not isinstance(msg, discord.Message):
                await ctx.send("Message répondu inaccessible")
                return
            modifs = f"{ref} {modifs}"

        else:
            if (chan := tools.channel(ref, must_be_found=False)):
                msg_id = chan.last_message_id
            elif (mtch := re.fullmatch(r"https*://discord.com/channels/"
                                       r"(\d{18})/(\d{18})/(\d{18})/*", ref)):
                chan = ctx.guild.get_channel(int(mtch.group(2)))
                msg_id = mtch.group(3)
            elif (mtch := re.fullmatch(r"(\d{18})-(\d{18})", ref)):
                chan = ctx.guild.get_channel(int(mtch.group(1)))
                msg_id = mtch.group(2)
            elif (mtch := re.fullmatch(r"\d{18}", ref)):
                chan = ctx.channel
                msg_id = mtch.group(0)
            else:
                await ctx.send("Je dois modifier quoi là ??????????")
                return

            if not chan:
                await ctx.send("Channel spécifié introuvable")
                return

            try:
                msg = await chan.fetch_message(msg_id)
            except discord.errors.NotFound:
                await ctx.send("Pas trouvé, pas de chance")
                return

        if msg.author != ctx.bot.user:
           await ctx.send("Message ciblé non rédigé par le bot")
           return

        old = msg.content
        if not ">" in modifs:
            await ctx.send("L'ordre de remplacement doit contenir '>'.")
            return

        before, _, after = modifs.partition(">")
        if before:
            new = old.replace(before.strip(), after.strip())
        else:
            new = after.strip()

        if new == old:
            await ctx.send("Pas de remplacement à effectuer.")
            return

        await msg.edit(content=new)
        await ctx.send("Fait.")
